"""CellDesigner XML parsing utilities.

Pure XML traversal and data-extraction functions for CellDesigner documents.
No momapy object construction happens here: all functions work with lxml
objectify trees and return raw Python / geometry values.
"""

import collections

import momapy.geometry
import momapy.sbml.io.sbml._parsing

_CD_NAMESPACE = "http://www.sbml.org/2001/ns/celldesigner"

_LINK_ANCHOR_POSITION_TO_ANCHOR_NAME = {
    "NW": "north_west",
    "NNW": "north_north_west",
    "N": "north",
    "NNE": "north_north_east",
    "NE": "north_east",
    "ENE": "east_north_east",
    "E": "east",
    "ESE": "east_south_east",
    "SE": "south_east",
    "SSE": "south_south_east",
    "S": "south",
    "SSW": "south_south_west",
    "SW": "south_west",
    "WSW": "west_south_west",
    "W": "west",
    "WNW": "west_north_west",
}

_TEXT_TO_CHARACTER = {
    "_underscore_": "_",
    "_br_": "\n",
    "_BR_": "\n",
    "_plus_": "+",
    "_minus_": "-",
    "_slash_": "/",
    "_space_": " ",
    "_Alpha_": "Α",
    "_alpha_": "α",
    "_Beta_": "Β",
    "_beta_": "β",
    "_Gamma_": "Γ",
    "_gamma_": "γ",
    "_Delta_": "Δ",
    "_delta_": "δ",
    "_Epsilon_": "Ε",
    "_epsilon_": "ε",
    "_Zeta_": "Ζ",
    "_zeta_": "ζ",
    "_Eta_": "Η",
    "_eta_": "η",
    "_Theta_": "Θ",
    "_theta_": "θ",
    "_Iota_": "Ι",
    "_iota_": "ι",
    "_Kappa_": "Κ",
    "_kappa_": "κ",
    "_Lambda_": "Λ",
    "_lambda_": "λ",
    "_Mu_": "Μ",
    "_mu_": "μ",
    "_Nu_": "Ν",
    "_nu_": "ν",
    "_Xi_": "Ξ",
    "_xi_": "ξ",
    "_Omicron_": "Ο",
    "_omicron_": "ο",
    "_Pi_": "Π",
    "_pi_": "π",
    "_Rho_": "Ρ",
    "_rho_": "ρ",
    "_Sigma_": "Σ",
    "_sigma_": "σ",
    "_Tau_": "Τ",
    "_tau_": "τ",
    "_Upsilon_": "Υ",
    "_upsilon_": "υ",
    "_Phi_": "Φ",
    "_phi_": "φ",
    "_Chi_": "Χ",
    "_chi_": "χ",
    "_Psi_": "Ψ",
    "_psi_": "ψ",
    "_Omega_": "Ω",
    "_omega_": "ω",
}


def get_annotation(cd_element):
    return getattr(cd_element, "annotation", None)


def get_extension(cd_element):
    cd_annotation = get_annotation(cd_element)
    if cd_annotation is None:
        return None
    cd_extension = getattr(
        cd_element.annotation, f"{{{_CD_NAMESPACE}}}extension", None
    )
    return cd_extension


def get_species(cd_model):
    list_of_species = getattr(cd_model, "listOfSpecies", None)
    if list_of_species is None:
        return []
    return list(getattr(list_of_species, "species", []))


def get_reactions(cd_model):
    list_of_reactions = getattr(cd_model, "listOfReactions", None)
    if list_of_reactions is None:
        return []
    return list(getattr(list_of_reactions, "reaction", []))


def get_species_aliases(cd_model):
    # only the non-included ones
    extension = get_extension(cd_model)
    species_aliases = list(
        getattr(extension.listOfSpeciesAliases, "speciesAlias", [])
    )
    species_aliases = [
        species_alias
        for species_alias in species_aliases
        if species_alias.get("complexSpeciesAlias") is None
    ]
    return species_aliases


def get_included_species_aliases(cd_model):
    extension = get_extension(cd_model)
    species_aliases = list(
        getattr(extension.listOfSpeciesAliases, "speciesAlias", [])
    )
    species_aliases = [
        species_alias
        for species_alias in species_aliases
        if species_alias.get("complexSpeciesAlias") is not None
    ]
    return species_aliases


def get_complex_species_aliases(cd_model):
    # only the non-included ones
    extension = get_extension(cd_model)
    complex_species_aliases = list(
        getattr(
            extension.listOfComplexSpeciesAliases,
            "complexSpeciesAlias",
            [],
        )
    )
    complex_species_aliases = [
        complex_species_alias
        for complex_species_alias in complex_species_aliases
        if complex_species_alias.get("complexSpeciesAlias") is None
    ]
    return complex_species_aliases


def get_included_complex_species_aliases(cd_model):
    extension = get_extension(cd_model)
    complex_species_aliases = list(
        getattr(
            extension.listOfComplexSpeciesAliases,
            "complexSpeciesAlias",
            [],
        )
    )
    complex_species_aliases = [
        complex_species_alias
        for complex_species_alias in complex_species_aliases
        if complex_species_alias.get("complexSpeciesAlias") is not None
    ]
    return complex_species_aliases


def get_compartments(cd_model):
    return list(getattr(cd_model.listOfCompartments, "compartment", []))


def get_compartment_aliases(cd_model):
    extension = get_extension(cd_model)
    return list(getattr(extension.listOfCompartmentAliases, "compartmentAlias", []))


def get_protein_templates(cd_model):
    extension = get_extension(cd_model)
    return list(getattr(extension.listOfProteins, "protein", []))


def get_gene_templates(cd_model):
    extension = get_extension(cd_model)
    return list(getattr(extension.listOfGenes, "gene", []))


def get_rna_templates(cd_model):
    extension = get_extension(cd_model)
    return list(getattr(extension.listOfRNAs, "RNA", []))


def get_antisense_rna_templates(cd_model):
    extension = get_extension(cd_model)
    return list(getattr(extension.listOfAntisenseRNAs, "AntisenseRNA", []))


def get_species_templates(cd_model):
    return (
        get_protein_templates(cd_model)
        + get_gene_templates(cd_model)
        + get_rna_templates(cd_model)
        + get_antisense_rna_templates(cd_model)
    )


def get_modification_residues(cd_species_template):
    list_of_modification_residues = getattr(
        cd_species_template, "listOfModificationResidues", None
    )
    if list_of_modification_residues is not None:
        modification_residues = list(
            getattr(list_of_modification_residues, "modificationResidue", [])
        )
    else:
        modification_residues = []
    return modification_residues


def get_regions(cd_species_template):
    list_of_regions = getattr(cd_species_template, "listOfRegions", None)
    if list_of_regions is not None:
        regions = list(getattr(list_of_regions, "region", []))
    else:
        regions = []
    return regions


def get_included_species(cd_model):
    extension = get_extension(cd_model)
    list_of_included_species = getattr(extension, "listOfIncludedSpecies", None)
    if list_of_included_species is None:
        return []
    return list(getattr(list_of_included_species, "species", []))


def get_reaction_type(cd_reaction):
    cd_extension = get_extension(cd_reaction)
    return cd_extension.reactionType.text


def get_gate_members(cd_reaction):
    cd_extension = get_extension(cd_reaction)
    list_of_gate_member = getattr(cd_extension, "listOfGateMember", None)
    if list_of_gate_member is not None:
        gate_members = list(getattr(list_of_gate_member, "GateMember", []))
    else:
        gate_members = []
    return gate_members


def get_key_from_reaction(cd_reaction):
    cd_reaction_type = get_reaction_type(cd_reaction)
    if cd_reaction_type == "BOOLEAN_LOGIC_GATE":
        cd_gate_member = get_gate_members(cd_reaction)[0]
        cd_reaction_type = cd_gate_member.get("modificationType")
    return ("REACTION", cd_reaction_type)


def get_key_from_species_template(cd_species_template):
    return (
        "TEMPLATE",
        cd_species_template.get("type"),
    )


def get_key_from_region(cd_region):
    return (
        "REGION",
        cd_region.get("type"),
    )


def get_key_from_species(cd_species, cd_id_to_cd_element):
    cd_species_template = get_template_from_species(
        cd_species, cd_id_to_cd_element
    )
    if cd_species_template is None:
        key = get_class_from_species(cd_species).text
    else:
        key = cd_species_template.get("type")
    return ("SPECIES", key)


def get_key_from_reaction_modification(cd_reaction_modification):
    if has_boolean_input_from_modification(
        cd_reaction_modification
    ):
        cd_reaction_modification_type = cd_reaction_modification.get(
            "modificationType"
        )
    else:
        cd_reaction_modification_type = cd_reaction_modification.get("type")
    return ("MODIFIER", cd_reaction_modification_type)


def get_key_from_gate_member(cd_reaction_modification):
    return ("GATE", cd_reaction_modification.get("type"))


def get_identity(cd_species):
    cd_extension = get_extension(cd_species)
    if cd_extension is not None:
        cd_identity = cd_extension.speciesIdentity
    else:
        cd_identity = cd_species.annotation.speciesIdentity
    return cd_identity


def get_class_from_species(cd_species):
    cd_identity = get_identity(cd_species)
    if cd_identity is None:
        return None
    cd_class = getattr(cd_identity, "class", None)
    return cd_class


def get_state(cd_species):
    cd_species_identity = get_identity(cd_species)
    if cd_species_identity is None:
        return None
    cd_species_state = getattr(cd_species_identity, "state", None)
    return cd_species_state


def get_activity(cd_species_alias):
    return getattr(cd_species_alias, "activity", None)


def get_homodimer(cd_species):
    cd_species_state = get_state(cd_species)
    if cd_species_state is None:
        return None
    return getattr(cd_species_state, "homodimer", None)


def get_hypothetical(cd_species):
    cd_species_identity = get_identity(cd_species)
    if cd_species_identity is None:
        return None
    return getattr(cd_species_identity, "hypothetical", None)


def get_species_modifications(cd_species):
    cd_species_state = get_state(cd_species)
    if cd_species_state is None:
        return []
    cd_list_of_modifications = getattr(
        cd_species_state, "listOfModifications", None
    )
    if cd_list_of_modifications is None:
        return []
    return list(getattr(cd_list_of_modifications, "modification", []))


def get_species_structural_states(cd_species):
    cd_species_state = get_state(cd_species)
    if cd_species_state is None:
        return []
    cd_list_of_structural_states = getattr(
        cd_species_state, "listOfStructuralStates", None
    )
    if cd_list_of_structural_states is None:
        return []
    return cd_list_of_structural_states.structuralState


def get_template_from_species(cd_species, cd_id_to_cd_element):
    cd_species_identity = get_identity(cd_species)
    for cd_species_template_type in [
        "proteinReference",
        "rnaReference",
        "geneReference",
        "antisensernaReference",
    ]:
        cd_reference = getattr(cd_species_identity, cd_species_template_type, None)
        if cd_reference is not None:
            cd_species_template = cd_id_to_cd_element[cd_reference.text]
            return cd_species_template
    return None


def get_template_from_species_alias(cd_species_alias, cd_id_to_cd_element):
    cd_species_id = cd_species_alias.get("species")
    cd_species = cd_id_to_cd_element[cd_species_id]
    return get_template_from_species(cd_species, cd_id_to_cd_element)


def get_bounds(cd_element):
    cd_x = cd_element.bounds.get("x")
    cd_y = cd_element.bounds.get("y")
    cd_w = cd_element.bounds.get("w")
    cd_h = cd_element.bounds.get("h")
    return cd_x, cd_y, cd_w, cd_h


def get_anchor_name_for_frame(cd_element):
    if getattr(cd_element, "linkAnchor", None) is not None:
        cd_element_anchor = cd_element.linkAnchor.get("position")
        anchor_name = _LINK_ANCHOR_POSITION_TO_ANCHOR_NAME[cd_element_anchor]
    else:
        anchor_name = "center"
    return anchor_name


def get_reactants(cd_reaction):
    return list(getattr(cd_reaction.listOfReactants, "speciesReference", []))


def get_base_reactants(cd_reaction):
    extension = get_extension(cd_reaction)
    return list(getattr(extension.baseReactants, "baseReactant", []))


def get_reactant_links(cd_reaction):
    extension = get_extension(cd_reaction)
    list_of_reactant_links = getattr(extension, "listOfReactantLinks", None)
    if list_of_reactant_links is None:
        return []
    return list(getattr(list_of_reactant_links, "reactantLink", []))


def get_products(cd_reaction):
    return list(getattr(cd_reaction.listOfProducts, "speciesReference", []))


def get_base_products(cd_reaction):
    extension = get_extension(cd_reaction)
    return list(getattr(extension.baseProducts, "baseProduct", []))


def get_product_links(cd_reaction):
    extension = get_extension(cd_reaction)
    list_of_product_links = getattr(extension, "listOfProductLinks", None)
    if list_of_product_links is None:
        return []
    return list(getattr(list_of_product_links, "productLink", []))


def get_edit_points_from_reaction(cd_reaction):
    extension = get_extension(cd_reaction)
    return getattr(extension, "editPoints", None)


def get_edit_points_from_participant_link(cd_participant_link):
    return getattr(cd_participant_link, "editPoints", None)


def get_rectangle_index(cd_reaction):
    extension = get_extension(cd_reaction)
    return extension.connectScheme.get("rectangleIndex")


def get_t_shape_index(cd_reaction):
    cd_edit_points = get_edit_points_from_reaction(cd_reaction)
    return cd_edit_points.get("tShapeIndex")


def has_boolean_input_from_modification(cd_reaction_modification):
    return cd_reaction_modification.get("type") in [
        "BOOLEAN_LOGIC_GATE_AND",
        "BOOLEAN_LOGIC_GATE_OR",
        "BOOLEAN_LOGIC_GATE_NOT",
        "BOOLEAN_LOGIC_GATE_UNKNOWN",
    ]


def has_boolean_input_from_reaction(cd_reaction):
    return (
        get_reaction_type(cd_reaction) == "BOOLEAN_LOGIC_GATE"
    )


def get_reaction_modifications(cd_reaction):
    extension = get_extension(cd_reaction)
    list_of_modification = getattr(extension, "listOfModification", None)
    if list_of_modification is None:
        return []
    modifications = list(getattr(list_of_modification, "modification", []))
    true_modifications = []
    cd_input_ids = set([])
    for modification in modifications:
        if has_boolean_input_from_modification(modification):
            true_modifications.append(modification)
            cd_input_ids.update(modification.get("modifiers").split(","))
    for modification in modifications:
        if (
            modification not in true_modifications
            and modification.get("modifiers") not in cd_input_ids
        ):
            true_modifications.append(modification)
    return true_modifications


def get_width(cd_model):
    extension = get_extension(cd_model)
    return extension.modelDisplay.get("sizeX")


def get_height(cd_model):
    extension = get_extension(cd_model)
    return extension.modelDisplay.get("sizeY")


def make_name(name):
    if name is None:
        return name
    for s, char in _TEXT_TO_CHARACTER.items():
        name = name.replace(s, char)
    return name


def get_notes(cd_element):
    return getattr(cd_element, "notes", None)


def get_rdf_from_notes(cd_notes):
    rdfs = list(cd_notes.iter(f"{{{momapy.sbml.io.sbml._parsing._RDF_NAMESPACE}}}RDF"))
    if rdfs:
        return rdfs[0]
    return None


def get_rdf(cd_element):
    annotation = get_annotation(cd_element)
    if annotation is None:
        return None
    return getattr(annotation, f"{{{momapy.sbml.io.sbml._parsing._RDF_NAMESPACE}}}RDF", None)


def make_id_to_element_mapping(cd_model):
    cd_id_to_cd_element = {}
    # compartments
    for cd_compartment in get_compartments(cd_model):
        cd_id_to_cd_element[cd_compartment.get("id")] = cd_compartment
    # compartment aliases
    for cd_compartment_alias in get_compartment_aliases(cd_model):
        cd_id_to_cd_element[cd_compartment_alias.get("id")] = cd_compartment_alias
    # species templates
    for cd_species_template in get_species_templates(cd_model):
        cd_id_to_cd_element[cd_species_template.get("id")] = cd_species_template
        for cd_modification_residue in get_modification_residues(
            cd_species_template
        ):
            cd_modification_residue_id = f"{cd_species_template.get('id')}_{cd_modification_residue.get('id')}"
            cd_id_to_cd_element[cd_modification_residue_id] = (
                cd_modification_residue
            )
        for cd_region in get_regions(cd_species_template):
            cd_region_id = f"{cd_species_template.get('id')}_{cd_region.get('id')}"
            cd_id_to_cd_element[cd_region_id] = cd_region
    # species
    for cd_species in get_species(cd_model):
        cd_id_to_cd_element[cd_species.get("id")] = cd_species
    # included species
    for cd_included_species in get_included_species(cd_model):
        cd_id_to_cd_element[cd_included_species.get("id")] = cd_included_species
    # species aliases
    for cd_species_alias in get_species_aliases(cd_model):
        cd_id_to_cd_element[cd_species_alias.get("id")] = cd_species_alias
    # complex species aliases
    for cd_complex_species_alias in get_complex_species_aliases(cd_model):
        cd_id_to_cd_element[cd_complex_species_alias.get("id")] = (
            cd_complex_species_alias
        )
    # included species aliases
    for cd_species_alias in get_included_species_aliases(cd_model):
        cd_id_to_cd_element[cd_species_alias.get("id")] = cd_species_alias
    # included complex species aliases
    for cd_complex_species_alias in get_included_complex_species_aliases(cd_model):
        cd_id_to_cd_element[cd_complex_species_alias.get("id")] = (
            cd_complex_species_alias
        )
    return cd_id_to_cd_element


def make_complex_alias_to_included_ids_mapping(cd_model):
    cd_complex_alias_id_to_cd_included_species_ids = collections.defaultdict(list)
    for cd_species_alias in get_included_species_aliases(cd_model):
        cd_complex_species_alias_id = cd_species_alias.get("complexSpeciesAlias")
        if cd_complex_species_alias_id is not None:
            cd_complex_alias_id_to_cd_included_species_ids[
                cd_complex_species_alias_id
            ].append(cd_species_alias.get("id"))

    for cd_species_alias in get_included_complex_species_aliases(cd_model):
        cd_complex_species_alias_id = cd_species_alias.get("complexSpeciesAlias")
        if cd_complex_species_alias_id is not None:
            cd_complex_alias_id_to_cd_included_species_ids[
                cd_complex_species_alias_id
            ].append(cd_species_alias.get("id"))
    return cd_complex_alias_id_to_cd_included_species_ids


def get_ordered_compartment_aliases(cd_model, cd_id_to_cd_element):
    cd_compartments = get_compartments(cd_model)
    cd_compartment_id_to_cd_outside_id = {}
    for cd_compartment in cd_compartments:
        cd_compartment_id = cd_compartment.get("id")
        cd_outside_id = cd_compartment.get("outside")
        cd_compartment_id_to_cd_outside_id[cd_compartment_id] = cd_outside_id
    ordered_cd_compartments = []
    while cd_compartment_id_to_cd_outside_id:
        for (
            cd_compartment_id,
            cd_outside_id,
        ) in cd_compartment_id_to_cd_outside_id.items():
            if cd_outside_id is None:
                ordered_cd_compartments.append(cd_compartment_id)
                to_delete = cd_compartment_id
                break
        del cd_compartment_id_to_cd_outside_id[to_delete]
        for (
            cd_compartment_id,
            cd_outside_id,
        ) in cd_compartment_id_to_cd_outside_id.items():
            if cd_outside_id == to_delete:
                cd_compartment_id_to_cd_outside_id[cd_compartment_id] = None
    cd_compartment_aliases = get_compartment_aliases(cd_model)
    ordered_cd_compartment_aliases = sorted(
        cd_compartment_aliases,
        key=lambda cd_compartment_alias: ordered_cd_compartments.index(
            cd_id_to_cd_element[cd_compartment_alias.get("compartment")].get("id")
        ),
    )
    return ordered_cd_compartment_aliases



"""Shared pytest fixtures for momapy tests."""

import pytest
import momapy.geometry
import momapy.coloring
import momapy.core
import momapy.core.layout
import momapy.drawing


@pytest.fixture
def sample_point():
    """Create a sample Point for testing."""
    return momapy.geometry.Point(10.0, 20.0)


@pytest.fixture
def sample_bbox():
    """Create a sample BBox for testing."""
    return momapy.geometry.Bbox(
        momapy.geometry.Point(0.0, 0.0), width=100.0, height=50.0
    )


@pytest.fixture
def sample_color():
    """Create a sample color for testing."""
    return momapy.coloring.black


@pytest.fixture
def sample_layout():
    """Create a simple Layout for testing."""
    return momapy.core.layout.Layout(
        position=momapy.geometry.Point(0, 0), width=200, height=200, layout_elements=[]
    )


@pytest.fixture
def sample_node():
    """Create a sample Rectangle node for testing."""
    import momapy.meta.nodes

    return momapy.meta.nodes.Rectangle(
        position=momapy.geometry.Point(50.0, 50.0), width=100.0, height=60.0
    )


@pytest.fixture
def sample_path():
    """Create a sample Path with common actions for testing."""
    return momapy.drawing.Path(
        actions=(
            momapy.drawing.MoveTo(momapy.geometry.Point(0, 0)),
            momapy.drawing.LineTo(momapy.geometry.Point(10, 0)),
            momapy.drawing.LineTo(momapy.geometry.Point(10, 10)),
            momapy.drawing.LineTo(momapy.geometry.Point(0, 10)),
            momapy.drawing.ClosePath(),
        )
    )


@pytest.fixture(scope="session")
def sample_sbgn_map():
    """Load a sample SBGN map for testing (session-scoped to avoid repeated I/O)."""
    import momapy.io
    import pathlib

    # Find a test SBGN file
    test_dir = pathlib.Path(__file__).parent
    sbgn_test_files = list(test_dir.glob("sbgn/io/test_files/*.sbgn"))

    if sbgn_test_files:
        return momapy.io.core.read(sbgn_test_files[0])
    else:
        # Return None if no test files found, tests should skip
        return None



"""Rendering subpackage for exporting maps to various formats.

Provides functions and registries for rendering molecular maps to
images (PNG, PDF, SVG) using different backends (Skia, Cairo, SVG).

Example:
    >>> from momapy.rendering import get_renderer, list_renderers
    >>> renderer = get_renderer("skia")
    >>> list_renderers()
    ['cairo', 'skia', 'svg-native']
"""

from __future__ import annotations

import typing

import momapy.plugins.core

if typing.TYPE_CHECKING:
    import momapy.rendering.core


renderer_registry = momapy.plugins.core.PluginRegistry(
    entry_point_group="momapy.renderers",
)


def get_renderer(name: str) -> type[momapy.rendering.core.Renderer]:
    """Get a renderer class by name.

    Args:
        name: Renderer name (e.g., "skia", "cairo", "svg-native").

    Returns:
        Renderer class for the specified backend.

    Raises:
        ValueError: If no renderer with that name exists.

    Example:
        >>> from momapy.rendering import get_renderer
        >>> renderer = get_renderer("skia")
    """
    renderer = renderer_registry.get(name)
    if renderer is None:
        available = renderer_registry.list_available()
        raise ValueError(
            f"No renderer named '{name}'. Available renderers: {', '.join(available)}"
        )
    return renderer


def list_renderers() -> list[str]:
    """List all available renderer names.

    Returns:
        Sorted list of available renderer names.

    Example:
        >>> from momapy.rendering import list_renderers
        >>> list_renderers()
        ['cairo', 'skia', 'svg-native']
    """
    return renderer_registry.list_available()


def register_renderer(name: str, cls: type[momapy.rendering.core.Renderer]) -> None:
    """Register a renderer class.

    Args:
        name: Name to register the renderer under.
        cls: Renderer class (must inherit from Renderer).

    Example:
        >>> from momapy.rendering import register_renderer
        >>> register_renderer("myrenderer", MyRenderer)
    """
    renderer_registry.register(name, cls)


def register_lazy_renderer(name: str, import_path: str) -> None:
    """Register a renderer for lazy loading.

    Args:
        name: Name to register the renderer under.
        import_path: Import path in format "module.path:ClassName".

    Example:
        >>> from momapy.rendering import register_lazy_renderer
        >>> register_lazy_renderer("myrenderer", "mymodule.rendering:MyRenderer")
    """
    renderer_registry.register_lazy(name, import_path)


for name, import_path in [
    ("svg-native", "momapy.rendering.svg_native:SVGNativeRenderer"),
    ("svg-native-compat", "momapy.rendering.svg_native:SVGNativeCompatRenderer"),
    ("skia", "momapy.rendering.skia:SkiaRenderer"),
    ("cairo", "momapy.rendering.cairo:CairoRenderer"),
]:
    register_lazy_renderer(name, import_path)



"""Top-level Map class."""

import dataclasses

import momapy.core.elements
import momapy.core.layout
import momapy.core.model
import momapy.core.mapping


@dataclasses.dataclass(frozen=True, kw_only=True)
class Map(momapy.core.elements.MapElement):
    """Class for maps"""

    model: momapy.core.model.Model | None = dataclasses.field(
        default=None, metadata={"description": "The model of the map"}
    )
    layout: momapy.core.layout.Layout | None = dataclasses.field(
        default=None, metadata={"description": "The layout of the map"}
    )
    layout_model_mapping: momapy.core.mapping.LayoutModelMapping | None = dataclasses.field(
        default=None,
        metadata={"description": "The layout model mapping of the map"},
    )

    def is_submap(self, other):
        """Return `true` if another given map is a submap of the `Map`, `false` otherwise"""
        if (
            self.model is None
            or self.layout is None
            or self.layout_model_mapping is None
        ):
            return False
        return (
            self.model.is_submodel(other.model)
            and self.layout.is_sublayout(other.layout)
            and self.layout_model_mapping.is_submapping(other.layout_model_mapping)
        )

    def get_mapping(
        self,
        map_element: "momapy.core.elements.MapElement | tuple",
    ):
        """Return the layout elements mapped to the given model element"""
        return self.layout_model_mapping.get_mapping(map_element)



"""Shared fixtures for rendering tests."""
import pytest
import tempfile
import os
import momapy.io.core


# Get the directory containing this conftest file
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
SBGN_MAPS_DIR = os.path.join(TEST_DIR, "..", "sbgn", "maps", "pd")


def get_sbgn_files():
    """Get all SBGN files from the maps directory."""
    if not os.path.exists(SBGN_MAPS_DIR):
        return []
    return [f for f in os.listdir(SBGN_MAPS_DIR) if f.endswith('.sbgn')]


SBGN_FILES = get_sbgn_files()


@pytest.fixture
def temp_dir():
    """Create a temporary directory for test outputs."""
    with tempfile.TemporaryDirectory() as tmpdir:
        yield tmpdir


@pytest.fixture
def sample_map():
    """Load a sample SBGN map for rendering tests."""
    if not SBGN_FILES:
        pytest.skip("No SBGN files found for testing")
    # Use the first available SBGN file
    input_file = os.path.join(SBGN_MAPS_DIR, SBGN_FILES[0])
    result = momapy.io.core.read(input_file)
    return result.obj



"""Tests for momapy.io.core module."""

import pytest
import momapy.io
import momapy.io.core


def test_reader_registry_exists():
    """Test that reader registry exists."""
    assert hasattr(momapy.io, "reader_registry")


def test_writer_registry_exists():
    """Test that writer registry exists."""
    assert hasattr(momapy.io, "writer_registry")


def test_register_reader():
    """Test register_reader function."""

    class DummyReader(momapy.io.core.Reader):
        @classmethod
        def read(cls, file_path, **options):
            return momapy.io.core.ReaderResult()

        @classmethod
        def check_file(cls, file_path):
            return False

    # Save original state
    original_readers = momapy.io.reader_registry.list_loaded()

    try:
        momapy.io.register_reader("test_reader", DummyReader)
        assert momapy.io.reader_registry.is_available("test_reader")
        assert momapy.io.reader_registry.get("test_reader") == DummyReader
    finally:
        # Restore original state - remove test reader
        if "test_reader" in momapy.io.reader_registry.list_loaded():
            del momapy.io.reader_registry._loaded_plugins["test_reader"]


def test_register_writer():
    """Test register_writer function."""

    class DummyWriter(momapy.io.core.Writer):
        @classmethod
        def write(cls, obj, file_path, **options):
            return momapy.io.core.WriterResult()

    # Save original state
    original_writers = momapy.io.writer_registry.list_loaded()

    try:
        momapy.io.register_writer("test_writer", DummyWriter)
        assert momapy.io.writer_registry.is_available("test_writer")
        assert momapy.io.writer_registry.get("test_writer") == DummyWriter
    finally:
        # Restore original state - remove test writer
        if "test_writer" in momapy.io.writer_registry.list_loaded():
            del momapy.io.writer_registry._loaded_plugins["test_writer"]


def test_io_result_creation():
    """Test IOResult creation."""
    result = momapy.io.core.IOResult()
    assert result.exceptions == []


def test_reader_result_creation():
    """Test ReaderResult creation."""
    result = momapy.io.core.ReaderResult()
    assert result.exceptions == []
    assert result.obj is None


def test_writer_result_creation():
    """Test WriterResult creation."""
    result = momapy.io.core.WriterResult()
    assert result.exceptions == []
    assert result.obj is None


def test_read_function_exists():
    """Test that read function exists."""
    assert hasattr(momapy.io.core, "read")
    assert callable(momapy.io.core.read)


def test_write_function_exists():
    """Test that write function exists."""
    assert hasattr(momapy.io.core, "write")
    assert callable(momapy.io.core.write)



"""Tests for momapy.geometry module."""

import pytest
import math
import numpy
import momapy.geometry


class TestPoint:
    """Tests for Point class."""

    def test_point_creation(self):
        """Test creating a Point."""
        p = momapy.geometry.Point(10.0, 20.0)
        assert p.x == 10.0
        assert p.y == 20.0

    def test_point_rounding(self):
        """Test that points are rounded on creation."""
        p = momapy.geometry.Point(10.123456789, 20.987654321)
        assert p.x == 10.1235  # ROUNDING = 4
        assert p.y == 20.9877

    def test_point_addition(self):
        """Test point addition."""
        p1 = momapy.geometry.Point(10.0, 20.0)
        p2 = momapy.geometry.Point(5.0, 3.0)
        result = p1 + p2
        assert result.x == 15.0
        assert result.y == 23.0

        # Test addition with tuple
        result2 = p1 + (5.0, 3.0)
        assert result2.x == 15.0
        assert result2.y == 23.0

    def test_point_subtraction(self):
        """Test point subtraction."""
        p1 = momapy.geometry.Point(10.0, 20.0)
        p2 = momapy.geometry.Point(5.0, 3.0)
        result = p1 - p2
        assert result.x == 5.0
        assert result.y == 17.0

    def test_point_multiplication(self):
        """Test point multiplication by scalar."""
        p = momapy.geometry.Point(10.0, 20.0)
        result = p * 2
        assert result.x == 20.0
        assert result.y == 40.0

    def test_point_division(self):
        """Test point division by scalar."""
        p = momapy.geometry.Point(10.0, 20.0)
        result = p / 2
        assert result.x == 5.0
        assert result.y == 10.0

    def test_point_iteration(self):
        """Test point iteration."""
        p = momapy.geometry.Point(10.0, 20.0)
        coords = list(p)
        assert coords == [10.0, 20.0]

    def test_point_to_tuple(self):
        """Test point to_tuple method."""
        p = momapy.geometry.Point(10.0, 20.0)
        t = p.to_tuple()
        assert t == (10.0, 20.0)

    def test_point_to_matrix(self):
        """Test point to_matrix method."""
        p = momapy.geometry.Point(10.0, 20.0)
        m = p.to_matrix()
        assert isinstance(m, numpy.ndarray)
        assert m[0][0] == 10.0
        assert m[1][0] == 20.0

    def test_point_from_tuple(self):
        """Test creating point from tuple."""
        p = momapy.geometry.Point.from_tuple((10.0, 20.0))
        assert p.x == 10.0
        assert p.y == 20.0

    def test_point_round(self):
        """Test point round method."""
        p = momapy.geometry.Point(10.123, 20.987)
        rounded = p.round(1)
        assert rounded.x == 10.1
        assert rounded.y == 21.0

    def test_point_isnan(self):
        """Test point isnan method."""
        p1 = momapy.geometry.Point(10.0, 20.0)
        assert not p1.isnan()

        p2 = momapy.geometry.Point(float("nan"), 20.0)
        assert p2.isnan()

    def test_point_bbox(self):
        """Test point bbox method."""
        p = momapy.geometry.Point(10.0, 20.0)
        bbox = p.bbox()
        assert isinstance(bbox, momapy.geometry.Bbox)
        assert bbox.width == 0
        assert bbox.height == 0


class TestLine:
    """Tests for Line class."""

    def test_line_creation(self):
        """Test creating a Line."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 10.0)
        line = momapy.geometry.Line(p1, p2)
        assert line.p1 == p1
        assert line.p2 == p2

    def test_line_slope(self):
        """Test line slope calculation."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 10.0)
        line = momapy.geometry.Line(p1, p2)
        assert line.slope() == 1.0

    def test_line_slope_vertical(self):
        """Test vertical line slope (should be NaN)."""
        p1 = momapy.geometry.Point(5.0, 0.0)
        p2 = momapy.geometry.Point(5.0, 10.0)
        line = momapy.geometry.Line(p1, p2)
        assert math.isnan(line.slope())

    def test_line_intercept(self):
        """Test line intercept calculation."""
        p1 = momapy.geometry.Point(0.0, 5.0)
        p2 = momapy.geometry.Point(10.0, 15.0)
        line = momapy.geometry.Line(p1, p2)
        assert line.intercept() == 5.0

    def test_line_intercept_rounded(self):
        """Test line intercept is rounded to ROUNDING decimals."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(3.0, 1.0)
        line = momapy.geometry.Line(p1, p2)
        intercept = line.intercept()
        assert intercept == round(intercept, momapy.geometry.ROUNDING)

    def test_line_slope_vertical(self):
        """Test line slope for vertical line is NaN."""
        p1 = momapy.geometry.Point(5.0, 0.0)
        p2 = momapy.geometry.Point(5.0, 10.0)
        line = momapy.geometry.Line(p1, p2)
        assert math.isnan(line.slope())


class TestSegment:
    """Tests for Segment class."""

    def test_segment_creation(self):
        """Test creating a Segment."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(3.0, 4.0)
        segment = momapy.geometry.Segment(p1, p2)
        assert segment.p1 == p1
        assert segment.p2 == p2

    def test_segment_length(self):
        """Test segment length calculation."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(3.0, 4.0)
        segment = momapy.geometry.Segment(p1, p2)
        assert segment.length() == 5.0

    def test_segment_get_angle_to_horizontal(self):
        """Test segment angle to horizontal."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 0.0)
        segment = momapy.geometry.Segment(p1, p2)
        assert segment.get_angle_to_horizontal() == pytest.approx(0.0, abs=0.01)

    def test_segment_bbox(self):
        """Test segment bbox method."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 10.0)
        segment = momapy.geometry.Segment(p1, p2)
        bbox = segment.bbox()
        assert isinstance(bbox, momapy.geometry.Bbox)


class TestEllipticalArc:
    """Tests for EllipticalArc class."""

    def test_elliptical_arc_creation(self):
        """Test creating an EllipticalArc."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 0.0)
        arc = momapy.geometry.EllipticalArc(
            p1, p2, rx=5.0, ry=5.0, x_axis_rotation=0.0, arc_flag=0, sweep_flag=1
        )
        assert arc.p1 == p1
        assert arc.p2 == p2
        assert arc.rx == 5.0
        assert arc.ry == 5.0

    def test_elliptical_arc_length(self):
        """Test elliptical arc length computation."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 0.0)
        arc = momapy.geometry.EllipticalArc(
            p1, p2, rx=5.0, ry=5.0, x_axis_rotation=0.0, arc_flag=0, sweep_flag=1
        )
        length = arc.length()
        # Semicircle with r=5 has length ~15.7
        assert length == pytest.approx(math.pi * 5, abs=0.5)


class TestBbox:
    """Tests for Bbox class."""

    def test_bbox_creation(self):
        """Test creating a Bbox."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        assert bbox.position == position
        assert bbox.width == 100.0
        assert bbox.height == 60.0

    def test_bbox_properties(self):
        """Test bbox x and y properties."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        assert bbox.x == 50.0
        assert bbox.y == 50.0

    def test_bbox_size(self):
        """Test bbox size method."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        size = bbox.size()
        assert size == (100.0, 60.0)

    def test_bbox_center(self):
        """Test bbox center anchor point."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        center = bbox.center()
        assert center.x == 50.0
        assert center.y == 50.0

    def test_bbox_north(self):
        """Test bbox north anchor point."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        north = bbox.north()
        assert north.x == 50.0
        assert north.y == 20.0  # 50 - 60/2

    def test_bbox_south(self):
        """Test bbox south anchor point."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        south = bbox.south()
        assert south.x == 50.0
        assert south.y == 80.0  # 50 + 60/2

    def test_bbox_east(self):
        """Test bbox east anchor point."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        east = bbox.east()
        assert east.x == 100.0  # 50 + 100/2
        assert east.y == 50.0

    def test_bbox_west(self):
        """Test bbox west anchor point."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        west = bbox.west()
        assert west.x == 0.0  # 50 - 100/2
        assert west.y == 50.0

    def test_bbox_north_east(self):
        """Test bbox north_east anchor point."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        ne = bbox.north_east()
        assert ne.x == 100.0
        assert ne.y == 20.0

    def test_bbox_isnan(self):
        """Test bbox isnan method."""
        position = momapy.geometry.Point(50.0, 50.0)
        bbox = momapy.geometry.Bbox(position, 100.0, 60.0)
        assert not bbox.isnan()

    def test_around_points_basic(self):
        """Test around_points with two points."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 20.0)
        bbox = momapy.geometry.Bbox.around_points([p1, p2])
        assert bbox.position == momapy.geometry.Point(5.0, 10.0)
        assert bbox.width == 10.0
        assert bbox.height == 20.0

    def test_around_points_single_point(self):
        """Test around_points with a single point gives zero-size bbox."""
        p = momapy.geometry.Point(3.0, 7.0)
        bbox = momapy.geometry.Bbox.around_points([p])
        assert bbox.position == p
        assert bbox.width == 0.0
        assert bbox.height == 0.0

    def test_around_points_empty_raises(self):
        """Test around_points raises ValueError on empty input."""
        with pytest.raises(ValueError):
            momapy.geometry.Bbox.around_points([])

    def test_around_points_collinear(self):
        """Test around_points with collinear horizontal points."""
        points = [
            momapy.geometry.Point(1.0, 5.0),
            momapy.geometry.Point(3.0, 5.0),
            momapy.geometry.Point(7.0, 5.0),
        ]
        bbox = momapy.geometry.Bbox.around_points(points)
        assert bbox.position == momapy.geometry.Point(4.0, 5.0)
        assert bbox.width == 6.0
        assert bbox.height == 0.0

    def test_around_points_generator(self):
        """Test around_points accepts a generator."""
        def gen():
            yield momapy.geometry.Point(0.0, 0.0)
            yield momapy.geometry.Point(4.0, 6.0)

        bbox = momapy.geometry.Bbox.around_points(gen())
        assert bbox.position == momapy.geometry.Point(2.0, 3.0)
        assert bbox.width == 4.0
        assert bbox.height == 6.0

    def test_union_empty(self):
        """Test union of empty list returns zero bbox."""
        bbox = momapy.geometry.Bbox.union([])
        assert bbox.position == momapy.geometry.Point(0.0, 0.0)
        assert bbox.width == 0.0
        assert bbox.height == 0.0

    def test_union_basic(self):
        """Test union of two bboxes."""
        b1 = momapy.geometry.Bbox(momapy.geometry.Point(5.0, 5.0), 10.0, 10.0)
        b2 = momapy.geometry.Bbox(momapy.geometry.Point(15.0, 15.0), 10.0, 10.0)
        bbox = momapy.geometry.Bbox.union([b1, b2])
        assert bbox.position == momapy.geometry.Point(10.0, 10.0)
        assert bbox.width == 20.0
        assert bbox.height == 20.0


class TestTransformations:
    """Tests for Transformation classes."""

    def test_translation_creation(self):
        """Test creating a Translation."""
        trans = momapy.geometry.Translation(10.0, 20.0)
        assert trans.tx == 10.0
        assert trans.ty == 20.0

    def test_translation_to_matrix(self):
        """Test translation to_matrix method."""
        trans = momapy.geometry.Translation(10.0, 20.0)
        m = trans.to_matrix()
        assert isinstance(m, numpy.ndarray)
        assert m[0][2] == 10.0
        assert m[1][2] == 20.0

    def test_rotation_creation(self):
        """Test creating a Rotation."""
        rot = momapy.geometry.Rotation(math.pi / 2)
        assert rot.angle == math.pi / 2

    def test_rotation_to_matrix(self):
        """Test rotation to_matrix method."""
        rot = momapy.geometry.Rotation(math.pi / 2)
        m = rot.to_matrix()
        assert isinstance(m, numpy.ndarray)

    def test_scaling_creation(self):
        """Test creating a Scaling."""
        scale = momapy.geometry.Scaling(2.0, 3.0)
        assert scale.sx == 2.0
        assert scale.sy == 3.0

    def test_scaling_to_matrix(self):
        """Test scaling to_matrix method."""
        scale = momapy.geometry.Scaling(2.0, 3.0)
        m = scale.to_matrix()
        assert isinstance(m, numpy.ndarray)
        assert m[0][0] == 2.0
        assert m[1][1] == 3.0

    def test_transform_point(self):
        """Test transforming a point with translation."""
        p = momapy.geometry.Point(10.0, 20.0)
        trans = momapy.geometry.Translation(5.0, 10.0)
        transformed = p.transformed(trans)
        assert transformed.x == 15.0
        assert transformed.y == 30.0

    def test_point_transformed_method(self):
        """Test point transformed method."""
        p = momapy.geometry.Point(10.0, 20.0)
        trans = momapy.geometry.Translation(5.0, 10.0)
        transformed = p.transformed(trans)
        assert transformed.x == 15.0
        assert transformed.y == 30.0

    def test_transformation_multiplication(self):
        """Test transformation multiplication."""
        trans1 = momapy.geometry.Translation(10.0, 20.0)
        trans2 = momapy.geometry.Translation(5.0, 10.0)
        combined = trans1 * trans2
        assert isinstance(combined, momapy.geometry.MatrixTransformation)

    def test_invert_scaling_reciprocal(self):
        """Test that inverted scaling returns reciprocal, not negation."""
        scaling = momapy.geometry.Scaling(2.0, 3.0)
        inverted = scaling.inverted()
        assert inverted.sx == 0.5  # 1/2, not -2
        assert inverted.sy == pytest.approx(1.0 / 3.0)  # 1/3, not -3

    def test_invert_scaling_composition(self):
        """Test that scaling and inverse return to original."""
        p = momapy.geometry.Point(10.0, 20.0)
        scaling = momapy.geometry.Scaling(2.0, 4.0)
        inverted = scaling.inverted()

        # Apply scaling then inverse
        scaled = p.transformed(scaling)
        restored = scaled.transformed(inverted)

        assert restored.x == pytest.approx(p.x, rel=1e-10)
        assert restored.y == pytest.approx(p.y, rel=1e-10)


class TestIntersectionFunctions:
    """Tests for intersection functions."""

    def test_intersection_of_perpendicular_lines(self):
        """Test intersection of perpendicular lines."""
        # Horizontal line: y = 5
        line1 = momapy.geometry.Line(
            momapy.geometry.Point(0.0, 5.0), momapy.geometry.Point(10.0, 5.0)
        )
        # Vertical line: x = 3
        line2 = momapy.geometry.Line(
            momapy.geometry.Point(3.0, 0.0), momapy.geometry.Point(3.0, 10.0)
        )

        intersections = line1.get_intersection_with_line(line2)
        assert len(intersections) == 1
        assert isinstance(intersections[0], momapy.geometry.Point)
        assert intersections[0].x == 3.0
        assert intersections[0].y == 5.0

    def test_intersection_of_parallel_lines(self):
        """Test that parallel lines return empty list."""
        line1 = momapy.geometry.Line(
            momapy.geometry.Point(0.0, 0.0), momapy.geometry.Point(10.0, 0.0)
        )
        line2 = momapy.geometry.Line(
            momapy.geometry.Point(0.0, 5.0), momapy.geometry.Point(10.0, 5.0)
        )

        intersections = line1.get_intersection_with_line(line2)
        assert len(intersections) == 0

    def test_intersection_line_and_segment_inside(self):
        """Test intersection when line crosses segment."""
        line = momapy.geometry.Line(
            momapy.geometry.Point(5.0, 0.0), momapy.geometry.Point(5.0, 10.0)
        )
        segment = momapy.geometry.Segment(
            momapy.geometry.Point(0.0, 5.0), momapy.geometry.Point(10.0, 5.0)
        )

        intersections = segment.get_intersection_with_line(line)
        assert len(intersections) == 1
        assert isinstance(intersections[0], momapy.geometry.Point)
        assert intersections[0].x == 5.0
        assert intersections[0].y == 5.0

    def test_intersection_line_and_segment_outside(self):
        """Test intersection when line misses segment."""
        line = momapy.geometry.Line(
            momapy.geometry.Point(15.0, 0.0), momapy.geometry.Point(15.0, 10.0)
        )
        segment = momapy.geometry.Segment(
            momapy.geometry.Point(0.0, 5.0), momapy.geometry.Point(10.0, 5.0)
        )

        intersections = segment.get_intersection_with_line(line)
        assert len(intersections) == 0


class TestCubicBezierCurve:
    """Tests for CubicBezierCurve class."""

    def test_cubic_bezier_curve_creation(self):
        """Test creating a CubicBezierCurve."""
        curve = momapy.geometry.CubicBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(3.0, 0.0),
            momapy.geometry.Point(1.0, 0.5),
            momapy.geometry.Point(2.0, 0.5),
        )
        assert curve.p1.x == 0.0
        assert curve.p2.x == 3.0

    def test_cubic_bezier_curve_evaluate_start(self):
        """Test evaluate at t=0 returns start point."""
        start = momapy.geometry.Point(0.0, 0.0)
        curve = momapy.geometry.CubicBezierCurve(
            start,
            momapy.geometry.Point(3.0, 0.0),
            momapy.geometry.Point(1.0, 1.0),
            momapy.geometry.Point(2.0, 1.0),
        )
        result = curve.evaluate(0.0)
        assert result.x == start.x
        assert result.y == start.y

    def test_cubic_bezier_curve_evaluate_end(self):
        """Test evaluate at t=1 returns end point."""
        end = momapy.geometry.Point(3.0, 0.0)
        curve = momapy.geometry.CubicBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            end,
            momapy.geometry.Point(1.0, 1.0),
            momapy.geometry.Point(2.0, 1.0),
        )
        result = curve.evaluate(1.0)
        assert result.x == end.x
        assert result.y == end.y

    def test_cubic_bezier_curve_evaluate_multi(self):
        """Test that cubic bezier curve evaluates at multiple points."""
        curve = momapy.geometry.CubicBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(3.0, 0.0),
            momapy.geometry.Point(1.0, 1.0),
            momapy.geometry.Point(2.0, 1.0),
        )
        import numpy
        points = curve.evaluate_multi(numpy.array([0.0, 0.5, 1.0]))
        assert len(points) == 3
        assert points[0].x == pytest.approx(0.0, abs=0.01)
        assert points[2].x == pytest.approx(3.0, abs=0.01)

    def test_cubic_bezier_curve_transformed(self):
        """Test transformed cubic bezier curve."""
        curve = momapy.geometry.CubicBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(3.0, 0.0),
            momapy.geometry.Point(1.0, 1.0),
            momapy.geometry.Point(2.0, 1.0),
        )
        translation = momapy.geometry.Translation(5.0, 10.0)
        transformed = curve.transformed(translation)
        assert transformed.p1.x == 5.0
        assert transformed.p1.y == 10.0
        assert transformed.p2.x == 8.0
        assert transformed.p2.y == 10.0

    def test_cubic_bezier_curve_reversed(self):
        """Test reversed cubic bezier curve swaps endpoints."""
        start = momapy.geometry.Point(0.0, 0.0)
        end = momapy.geometry.Point(3.0, 0.0)
        curve = momapy.geometry.CubicBezierCurve(
            start,
            end,
            momapy.geometry.Point(1.0, 1.0),
            momapy.geometry.Point(2.0, 1.0),
        )
        reversed_curve = curve.reversed()
        assert reversed_curve.p1.x == end.x
        assert reversed_curve.p2.x == start.x

    def test_cubic_bezier_curve_bbox(self):
        """Test bounding box of cubic bezier curve."""
        curve = momapy.geometry.CubicBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(3.0, 5.0),
            momapy.geometry.Point(7.0, 5.0),
        )
        bbox = curve.bbox()
        assert bbox is not None
        assert bbox.width > 0
        assert bbox.height > 0

    def test_cubic_bezier_curve_length(self):
        """Test arc length of cubic bezier curve."""
        curve = momapy.geometry.CubicBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(3.0, 5.0),
            momapy.geometry.Point(7.0, 5.0),
        )
        length = curve.length()
        # Should be longer than straight distance of 10
        assert length > 10.0

    def test_cubic_bezier_curve_split(self):
        """Test splitting cubic bezier curve."""
        curve = momapy.geometry.CubicBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(3.0, 5.0),
            momapy.geometry.Point(7.0, 5.0),
        )
        left, right = curve.split(0.5)
        assert left.p1.x == pytest.approx(0.0, abs=0.01)
        assert right.p2.x == pytest.approx(10.0, abs=0.01)


class TestQuadraticBezierCurve:
    """Tests for QuadraticBezierCurve class."""

    def test_quadratic_bezier_curve_creation(self):
        """Test creating a QuadraticBezierCurve."""
        curve = momapy.geometry.QuadraticBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(5.0, 5.0),
        )
        assert curve.p1.x == 0.0
        assert curve.p2.x == 10.0

    def test_quadratic_bezier_curve_evaluate_start(self):
        """Test evaluate at t=0 returns start point."""
        start = momapy.geometry.Point(0.0, 0.0)
        curve = momapy.geometry.QuadraticBezierCurve(
            start,
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(5.0, 5.0),
        )
        result = curve.evaluate(0.0)
        assert result.x == start.x
        assert result.y == start.y

    def test_quadratic_bezier_curve_evaluate_end(self):
        """Test evaluate at t=1 returns end point."""
        end = momapy.geometry.Point(10.0, 0.0)
        curve = momapy.geometry.QuadraticBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            end,
            momapy.geometry.Point(5.0, 5.0),
        )
        result = curve.evaluate(1.0)
        assert result.x == end.x
        assert result.y == end.y

    def test_quadratic_bezier_curve_evaluate_multi_with_list(self):
        """Test evaluate_multi accepts a plain list."""
        curve = momapy.geometry.QuadraticBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(5.0, 5.0),
        )
        results = curve.evaluate_multi([0.0, 0.5, 1.0])
        assert len(results) == 3
        assert results[0].x == 0.0
        assert results[2].x == 10.0

    def test_quadratic_bezier_curve_evaluate_delegates(self):
        """Test evaluate matches evaluate_multi for same parameter."""
        curve = momapy.geometry.QuadraticBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(5.0, 5.0),
        )
        single = curve.evaluate(0.5)
        multi = curve.evaluate_multi([0.5])[0]
        assert single.x == multi.x
        assert single.y == multi.y

    def test_quadratic_bezier_curve_length(self):
        """Test arc length of quadratic bezier curve."""
        curve = momapy.geometry.QuadraticBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(5.0, 5.0),
        )
        length = curve.length()
        assert length > 10.0


class TestSegmentMethods:
    """Tests for Segment additional methods."""

    def test_segment_get_position_at_fraction_start(self):
        """Test get_position_at_fraction at 0 returns p1."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 10.0)
        segment = momapy.geometry.Segment(p1, p2)
        result = segment.get_position_at_fraction(0.0)
        assert result.x == p1.x
        assert result.y == p1.y

    def test_segment_get_position_at_fraction_end(self):
        """Test get_position_at_fraction at 1 returns p2."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 10.0)
        segment = momapy.geometry.Segment(p1, p2)
        result = segment.get_position_at_fraction(1.0)
        assert result.x == p2.x
        assert result.y == p2.y

    def test_segment_get_position_at_fraction_midpoint(self):
        """Test get_position_at_fraction at 0.5 returns midpoint."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 10.0)
        segment = momapy.geometry.Segment(p1, p2)
        result = segment.get_position_at_fraction(0.5)
        assert result.x == 5.0
        assert result.y == 5.0

    def test_segment_shortened_from_start(self):
        """Test shortening segment from start."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 0.0)
        segment = momapy.geometry.Segment(p1, p2)
        shortened = segment.shortened(length=2.0, start_or_end="start")
        assert shortened.length() == pytest.approx(8.0)

    def test_segment_shortened_from_end(self):
        """Test shortening segment from end."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 0.0)
        segment = momapy.geometry.Segment(p1, p2)
        shortened = segment.shortened(length=2.0, start_or_end="end")
        assert shortened.length() == pytest.approx(8.0)

    def test_segment_reversed(self):
        """Test reversed segment swaps endpoints."""
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 10.0)
        segment = momapy.geometry.Segment(p1, p2)
        reversed_seg = segment.reversed()
        assert reversed_seg.p1.x == p2.x
        assert reversed_seg.p2.x == p1.x


class TestSegmentGetPositionAndAngleAtFraction:
    """Tests for Segment.get_position_and_angle_at_fraction."""

    def test_position_at_start(self):
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 0.0)
        segment = momapy.geometry.Segment(p1, p2)
        position, _ = segment.get_position_and_angle_at_fraction(0.0)
        assert position.x == pytest.approx(0.0, abs=0.01)
        assert position.y == pytest.approx(0.0, abs=0.01)

    def test_position_at_end(self):
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 0.0)
        segment = momapy.geometry.Segment(p1, p2)
        position, _ = segment.get_position_and_angle_at_fraction(1.0)
        assert position.x == pytest.approx(10.0, abs=0.01)
        assert position.y == pytest.approx(0.0, abs=0.01)

    def test_position_at_midpoint(self):
        p1 = momapy.geometry.Point(0.0, 0.0)
        p2 = momapy.geometry.Point(10.0, 0.0)
        segment = momapy.geometry.Segment(p1, p2)
        position, _ = segment.get_position_and_angle_at_fraction(0.5)
        assert position.x == pytest.approx(5.0, abs=0.01)
        assert position.y == pytest.approx(0.0, abs=0.01)

    def test_angle_horizontal_segment(self):
        segment = momapy.geometry.Segment(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
        )
        _, angle = segment.get_position_and_angle_at_fraction(0.5)
        assert angle == pytest.approx(0.0, abs=0.01)

    def test_angle_diagonal_segment(self):
        segment = momapy.geometry.Segment(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 10.0),
        )
        _, angle = segment.get_position_and_angle_at_fraction(0.5)
        assert angle == pytest.approx(math.pi / 4, abs=0.01)


class TestBezierCurveGetPositionAndAngleAtFraction:
    """Tests for CubicBezierCurve.get_position_and_angle_at_fraction.

    Uses a symmetric cubic curve (p1=(0,0), p2=(10,0), cp1=(3,5), cp2=(7,5))
    whose peak is at the arc-length midpoint (by symmetry).
    """

    @pytest.fixture
    def symmetric_curve(self):
        return momapy.geometry.CubicBezierCurve(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            momapy.geometry.Point(3.0, 5.0),
            momapy.geometry.Point(7.0, 5.0),
        )

    def test_position_at_start(self, symmetric_curve):
        position, _ = symmetric_curve.get_position_and_angle_at_fraction(0.0)
        assert position.x == pytest.approx(0.0, abs=0.01)
        assert position.y == pytest.approx(0.0, abs=0.01)

    def test_position_at_end(self, symmetric_curve):
        position, _ = symmetric_curve.get_position_and_angle_at_fraction(1.0)
        assert position.x == pytest.approx(10.0, abs=0.01)
        assert position.y == pytest.approx(0.0, abs=0.01)

    def test_angle_at_midpoint_is_horizontal(self, symmetric_curve):
        # At the arc-length midpoint of a symmetric curve the tangent is
        # horizontal, i.e. angle == 0 (mod 2π).  Before the _get_angle_at_fraction
        # fix (missing `previous_coord = current_coord` update) this returned
        # a wrong non-zero angle.
        _, angle = symmetric_curve.get_position_and_angle_at_fraction(0.5)
        # Normalise to (-π, π] for comparison
        angle_signed = (angle + math.pi) % (2 * math.pi) - math.pi
        assert angle_signed == pytest.approx(0.0, abs=0.05)


class TestEllipticalArcGetPositionAndAngleAtFraction:
    """Tests for EllipticalArc.get_position_and_angle_at_fraction."""

    @pytest.fixture
    def semicircle(self):
        # Semicircle: centre at (5,0), radius 5, sweeping upward
        return momapy.geometry.EllipticalArc(
            momapy.geometry.Point(0.0, 0.0),
            momapy.geometry.Point(10.0, 0.0),
            rx=5.0,
            ry=5.0,
            x_axis_rotation=0.0,
            arc_flag=0,
            sweep_flag=0,
        )

    def test_position_at_start(self, semicircle):
        position, _ = semicircle.get_position_and_angle_at_fraction(0.0)
        assert position.x == pytest.approx(0.0, abs=0.1)
        assert position.y == pytest.approx(0.0, abs=0.1)

    def test_position_at_end(self, semicircle):
        position, _ = semicircle.get_position_and_angle_at_fraction(1.0)
        assert position.x == pytest.approx(10.0, abs=0.1)
        assert position.y == pytest.approx(0.0, abs=0.1)

    def test_position_at_midpoint_is_bottom_of_arc(self, semicircle):
        # SVG y-axis points down; sweep_flag=0 sweeps downward, so the
        # midpoint of the semicircle is at (5, 5).
        position, _ = semicircle.get_position_and_angle_at_fraction(0.5)
        assert position.x == pytest.approx(5.0, abs=0.1)
        assert position.y == pytest.approx(5.0, abs=0.1)


class TestGeometryBorderAndAnchor:
    """Tests for geometry border and anchor functions."""

    def test_geometry_border(self):
        """Test getting border point of geometry primitives."""
        # Rectangle as 4 segments
        primitives = [
            momapy.geometry.Segment(
                momapy.geometry.Point(0.0, 0.0), momapy.geometry.Point(10.0, 0.0)
            ),
            momapy.geometry.Segment(
                momapy.geometry.Point(10.0, 0.0), momapy.geometry.Point(10.0, 10.0)
            ),
            momapy.geometry.Segment(
                momapy.geometry.Point(10.0, 10.0), momapy.geometry.Point(0.0, 10.0)
            ),
            momapy.geometry.Segment(
                momapy.geometry.Point(0.0, 10.0), momapy.geometry.Point(0.0, 0.0)
            ),
        ]
        south_point = momapy.geometry.Point(5.0, 20.0)
        border = momapy.geometry.get_primitives_border(primitives, south_point)
        assert border is not None
        assert isinstance(border, momapy.geometry.Point)

    def test_geometry_anchor_point(self):
        """Test getting anchor point from geometry primitives."""
        primitives = [
            momapy.geometry.Segment(
                momapy.geometry.Point(0.0, 0.0), momapy.geometry.Point(10.0, 0.0)
            ),
            momapy.geometry.Segment(
                momapy.geometry.Point(10.0, 0.0), momapy.geometry.Point(10.0, 10.0)
            ),
            momapy.geometry.Segment(
                momapy.geometry.Point(10.0, 10.0), momapy.geometry.Point(0.0, 10.0)
            ),
            momapy.geometry.Segment(
                momapy.geometry.Point(0.0, 10.0), momapy.geometry.Point(0.0, 0.0)
            ),
        ]
        anchor = momapy.geometry.get_primitives_anchor_point(primitives, "north")
        assert isinstance(anchor, momapy.geometry.Point)



"""Tests for momapy.positioning module."""

import pytest
import momapy.positioning
import momapy.geometry


def test_right_of_point():
    """Test right_of with a Point."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.right_of(point, 5.0)
    assert result.x == 15.0
    assert result.y == 20.0


def test_left_of_point():
    """Test left_of with a Point."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.left_of(point, 5.0)
    assert result.x == 5.0
    assert result.y == 20.0


def test_above_of_point():
    """Test above_of with a Point."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.above_of(point, 5.0)
    assert result.x == 10.0
    assert result.y == 15.0


def test_below_of_point():
    """Test below_of with a Point."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.below_of(point, 5.0)
    assert result.x == 10.0
    assert result.y == 25.0


def test_above_left_of_point():
    """Test above_left_of with a Point."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.above_left_of(point, 5.0, 3.0)
    assert result.x == 7.0  # 10 - 3
    assert result.y == 15.0  # 20 - 5


def test_above_left_of_point_single_distance():
    """Test above_left_of with single distance."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.above_left_of(point, 5.0)
    assert result.x == 5.0  # 10 - 5
    assert result.y == 15.0  # 20 - 5


def test_above_right_of_point():
    """Test above_right_of with a Point."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.above_right_of(point, 5.0, 3.0)
    assert result.x == 13.0  # 10 + 3
    assert result.y == 15.0  # 20 - 5


def test_below_left_of_point():
    """Test below_left_of with a Point."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.below_left_of(point, 5.0, 3.0)
    assert result.x == 7.0  # 10 - 3
    assert result.y == 25.0  # 20 + 5


def test_below_right_of_point():
    """Test below_right_of with a Point."""
    point = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.below_right_of(point, 5.0, 3.0)
    assert result.x == 13.0  # 10 + 3
    assert result.y == 25.0  # 20 + 5


def test_right_of_with_node():
    """Test right_of with a Node (via meta.nodes.Rectangle)."""
    import momapy.meta.nodes

    node = momapy.meta.nodes.Rectangle(
        position=momapy.geometry.Point(10.0, 20.0), width=20.0, height=10.0
    )
    result = momapy.positioning.right_of(node, 5.0)
    assert isinstance(result, momapy.geometry.Point)
    # Result should be to the right of the node's east anchor
    assert result.x > 10.0


def test_left_of_with_node():
    """Test left_of with a Node (via meta.nodes.Rectangle)."""
    import momapy.meta.nodes

    node = momapy.meta.nodes.Rectangle(
        position=momapy.geometry.Point(10.0, 20.0), width=20.0, height=10.0
    )
    result = momapy.positioning.left_of(node, 5.0)
    assert isinstance(result, momapy.geometry.Point)
    # Result should be to the left of the node's west anchor
    assert result.x < 10.0


def test_mid_of_two_points():
    """Test mid_of returns midpoint between two positions."""
    p1 = momapy.geometry.Point(0.0, 0.0)
    p2 = momapy.geometry.Point(10.0, 20.0)
    result = momapy.positioning.mid_of(p1, p2)
    assert result.x == 5.0
    assert result.y == 10.0


def test_fit_single_point():
    """Test fit with single point."""
    point = momapy.geometry.Point(10.0, 20.0)
    bbox = momapy.positioning.fit([point])
    assert isinstance(bbox, momapy.geometry.Bbox)
    assert bbox.position.x == 10.0
    assert bbox.position.y == 20.0


def test_fit_multiple_points():
    """Test fit with multiple points."""
    points = [
        momapy.geometry.Point(0.0, 0.0),
        momapy.geometry.Point(10.0, 20.0),
        momapy.geometry.Point(5.0, 15.0),
    ]
    bbox = momapy.positioning.fit(points)
    assert isinstance(bbox, momapy.geometry.Bbox)
    assert bbox.width > 0
    assert bbox.height > 0



"""Geometric primitives and transformations for momapy.

This module provides geometric classes and functions for working with points, lines,
segments, curves, and transformations. It includes support for Bezier curves,
elliptical arcs, and various geometric operations.

Examples:
    >>> from momapy.geometry import Point, Line, Segment, Rotation, Translation
    >>> # Create points
    >>> p1 = Point(0, 0)
    >>> p2 = Point(10, 10)
    >>> # Create a line
    >>> line = Line(p1, p2)
    >>> line.slope()
    1.0
    >>> # Create a segment
    >>> segment = Segment(p1, p2)
    >>> segment.length()
    14.14...
    >>> # Apply transformations
    >>> rotated = p1.transformed(Rotation(math.pi / 2, p2))
"""

import collections.abc
import dataclasses
import typing
import typing_extensions
import abc
import math
import copy

import numpy

import momapy.builder


ROUNDING = 4
ROUNDING_TOLERANCE = 10**-ROUNDING
ZERO_TOLERANCE = 1e-12
PARAMETER_TOLERANCE = 1e-10
CONVERGENCE_TOLERANCE = 1e-8


@dataclasses.dataclass(frozen=True)
class GeometryObject(abc.ABC):
    """Abstract base class for all geometry objects."""

    pass


@dataclasses.dataclass(frozen=True)
class Point(GeometryObject):
    """Represents a 2D point with x and y coordinates.

    Attributes:
        x: The x-coordinate.
        y: The y-coordinate.

    Examples:
        >>> p = Point(10, 20)
        >>> p.x
        10
        >>> p + (5, 5)
        Point(x=15, y=25)
    """

    x: float
    y: float

    def __post_init__(self):
        object.__setattr__(self, "x", round(self.x, ROUNDING))
        object.__setattr__(self, "y", round(self.y, ROUNDING))

    def __add__(self, xy):
        if momapy.builder.isinstance_or_builder(xy, Point):
            xy = (xy.x, xy.y)
        return Point(self.x + xy[0], self.y + xy[1])

    def __sub__(self, xy):
        if momapy.builder.isinstance_or_builder(xy, Point):
            xy = (xy.x, xy.y)
        return Point(self.x - xy[0], self.y - xy[1])

    def __mul__(self, scalar):
        return Point(self.x * scalar, self.y * scalar)

    def __truediv__(self, scalar):
        return Point(self.x / scalar, self.y / scalar)

    def __iter__(self):
        yield self.x
        yield self.y

    def to_matrix(self) -> numpy.ndarray:
        """Convert to a 3x1 numpy matrix for transformation operations.

        Returns:
            A numpy array [[x], [y], [1]].
        """
        m = numpy.array([[self.x], [self.y], [1]], dtype=float)
        return m

    def to_tuple(self) -> tuple[float, float]:
        """Convert to a tuple.

        Returns:
            Tuple (x, y).
        """
        return (
            self.x,
            self.y,
        )

    def get_intersection_with_line(self, line: "Line") -> list["Point"]:
        """Get intersection points with a line.

        Args:
            line: The line to intersect with.

        Returns:
            List of intersection points (empty if no intersection).
        """
        if line.has_point(self):
            return [self]
        return []

    def get_angle_to_horizontal(self) -> float:
        """Get the angle from origin to this point relative to horizontal.

        Returns:
            Angle in radians.
        """
        angle = math.atan2(self.y, self.x)
        return get_normalized_angle(angle)

    def transformed(self, transformation: "Transformation") -> "Point":
        """Apply a transformation to this point.

        Args:
            transformation: The transformation to apply.

        Returns:
            A new transformed Point.
        """
        m = numpy.matmul(transformation.to_matrix(), self.to_matrix())
        return Point(m[0][0], m[1][0])

    def reversed(self) -> "Point":
        """Return a reversed copy (identity for points).

        Returns:
            A copy of the point.
        """
        return Point(self.x, self.y)

    def round(self, ndigits=None):
        """Round coordinates to specified digits.

        Args:
            ndigits: Number of decimal places.

        Returns:
            A new Point with rounded coordinates.
        """
        return Point(round(self.x, ndigits), round(self.y, ndigits))

    def bbox(self) -> "Bbox":
        """Get the bounding box of this point.

        Returns:
            A Bbox with zero width and height.
        """
        return Bbox(copy.deepcopy(self), 0, 0)

    def isnan(self) -> bool:
        """Check if either coordinate is NaN.

        Returns:
            True if x or y is NaN.
        """
        return math.isnan(self.x) or math.isnan(self.y)

    @classmethod
    def from_tuple(cls, t: tuple[float, float]) -> typing_extensions.Self:
        """Create a Point from a tuple.

        Args:
            t: Tuple (x, y).

        Returns:
            A new Point.
        """
        return cls(t[0], t[1])


@dataclasses.dataclass(frozen=True)
class Line(GeometryObject):
    """Represents an infinite line defined by two points.

    Attributes:
        p1: First point on the line.
        p2: Second point on the line.

    Examples:
        >>> line = Line(Point(0, 0), Point(10, 10))
        >>> line.slope()
        1.0
        >>> line.intercept()
        0.0
    """

    p1: Point
    p2: Point

    def slope(self) -> float:
        """Calculate the slope of the line.

        Returns:
            The slope, or NaN if vertical.
        """
        if self.p1.x != self.p2.x:
            return round((self.p2.y - self.p1.y) / (self.p2.x - self.p1.x), ROUNDING)
        return float("NAN")

    def intercept(self) -> float:
        """Calculate the y-intercept of the line.

        Returns:
            The y-intercept, or NaN if vertical.
        """
        slope = self.slope()
        if not math.isnan(slope):
            return round(self.p1.y - slope * self.p1.x, ROUNDING)
        return float("NAN")

    def get_angle_to_horizontal(self) -> float:
        """Get the angle of the line relative to horizontal.

        Returns:
            Angle in radians.
        """
        angle = math.atan2(self.p2.y - self.p1.y, self.p2.x - self.p1.x)
        return get_normalized_angle(angle)

    def is_parallel_to_line(self, line: "Line") -> bool:
        """Check if this line is parallel to another.

        Args:
            line: The other line.

        Returns:
            True if parallel.
        """
        slope1 = self.slope()
        slope2 = line.slope()
        if math.isnan(slope1) and math.isnan(slope2):
            return True
        return slope1 == slope2

    def is_coincident_to_line(self, line: "Line") -> bool:
        """Check if this line is coincident with another.

        Args:
            line: The other line.

        Returns:
            True if coincident (same infinite line).
        """
        slope1 = self.slope()
        intercept1 = self.intercept()
        slope2 = line.slope()
        intercept2 = line.intercept()
        return (
            math.isnan(slope1)
            and math.isnan(slope2)
            and self.p1.x == line.p1.x
            or slope1 == slope2
            and intercept1 == intercept2
        )

    def get_intersection_with_line(self, line: "Line") -> list["Line"] | list[Point]:
        """Get intersection with another line.

        Args:
            line: The other line.

        Returns:
            List containing intersection point(s) or the coincident line.
        """
        slope1 = self.slope()
        intercept1 = self.intercept()
        slope2 = line.slope()
        intercept2 = line.intercept()
        if self.is_coincident_to_line(line):
            intersection = [copy.deepcopy(self)]
        elif self.is_parallel_to_line(line):
            intersection = []
        elif math.isnan(slope1):
            intersection = [Point(self.p1.x, slope2 * self.p1.x + intercept2)]
        elif math.isnan(slope2):
            intersection = [Point(line.p1.x, slope1 * line.p1.x + intercept1)]
        else:
            d = (self.p1.x - self.p2.x) * (line.p1.y - line.p2.y) - (
                self.p1.y - self.p2.y
            ) * (line.p1.x - line.p2.x)
            px = (
                (self.p1.x * self.p2.y - self.p1.y * self.p2.x)
                * (line.p1.x - line.p2.x)
                - (self.p1.x - self.p2.x)
                * (line.p1.x * line.p2.y - line.p1.y * line.p2.x)
            ) / d
            py = (
                (self.p1.x * self.p2.y - self.p1.y * self.p2.x)
                * (line.p1.y - line.p2.y)
                - (self.p1.y - self.p2.y)
                * (line.p1.x * line.p2.y - line.p1.y * line.p2.x)
            ) / d
            intersection = [Point(px, py)]
        return intersection

    def get_distance_to_point(self, point: Point) -> float:
        """Get perpendicular distance from a point to this line.

        Args:
            point: The point to measure from.

        Returns:
            The perpendicular distance.
        """
        return abs(
            (self.p2.x - self.p1.x) * (self.p1.y - point.y)
            - (self.p1.x - point.x) * (self.p2.y - self.p1.y)
        ) / math.sqrt((self.p2.x - self.p1.x) ** 2 + (self.p2.y - self.p1.y) ** 2)

    def has_point(self, point: Point, max_distance: float = ROUNDING_TOLERANCE) -> bool:
        """Check if a point lies on this line.

        Args:
            point: The point to check.
            max_distance: Maximum allowed distance from line.

        Returns:
            True if point is on the line within tolerance.
        """
        return self.get_distance_to_point(point) <= max_distance

    def transformed(self, transformation: "Transformation") -> "Line":
        """Apply a transformation to this line.

        Args:
            transformation: The transformation to apply.

        Returns:
            A new transformed Line.
        """
        return Line(
            self.p1.transformed(transformation),
            self.p2.transformed(transformation),
        )

    def reversed(self) -> "Line":
        """Return a reversed copy of the line.

        Returns:
            A new Line with p1 and p2 swapped.
        """
        return Line(self.p2, self.p1)


@dataclasses.dataclass(frozen=True)
class Segment(GeometryObject):
    """Represents a line segment between two points.

    Attributes:
        p1: Start point.
        p2: End point.

    Examples:
        >>> seg = Segment(Point(0, 0), Point(10, 10))
        >>> seg.length()
        14.14...
        >>> seg.get_position_at_fraction(0.5)
        Point(x=5.0, y=5.0)
    """

    p1: Point
    p2: Point

    def length(self) -> float:
        """Calculate the length of the segment.

        Returns:
            The Euclidean length.
        """
        return math.sqrt((self.p2.x - self.p1.x) ** 2 + (self.p2.y - self.p1.y) ** 2)

    def get_distance_to_point(self, point: Point) -> float:
        """Get shortest distance from a point to this segment.

        Args:
            point: The point to measure from.

        Returns:
            The shortest distance.
        """
        a = point.x - self.p1.x
        b = point.y - self.p1.y
        c = self.p2.x - self.p1.x
        d = self.p2.y - self.p1.y
        dot = a * c + b * d
        len_sq = c**2 + d**2
        if len_sq != 0:
            param = dot / len_sq
        else:
            param = -1
        if param < 0:
            xx = self.p1.x
            yy = self.p1.y
        elif param > 1:
            xx = self.p2.x
            yy = self.p2.y
        else:
            xx = self.p1.x + param * c
            yy = self.p1.y + param * d
        dx = point.x - xx
        dy = point.y - yy
        return math.sqrt(dx**2 + dy**2)

    def has_point(self, point: Point, max_distance: float = ROUNDING_TOLERANCE) -> bool:
        """Check if a point lies on this segment.

        Args:
            point: The point to check.
            max_distance: Maximum allowed distance.

        Returns:
            True if point is on the segment within tolerance.
        """
        return self.get_distance_to_point(point) <= max_distance

    def get_angle_to_horizontal(self) -> float:
        """Get the angle of the segment relative to horizontal.

        Returns:
            Angle in radians.
        """
        angle = math.atan2(self.p2.y - self.p1.y, self.p2.x - self.p1.x)
        return get_normalized_angle(angle)

    def get_intersection_with_line(self, line: Line) -> list[Point] | list["Segment"]:
        """Get intersection with a line.

        Args:
            line: The line to intersect with.

        Returns:
            List of intersection points or segment if coincident.
        """
        line2 = Line(self.p1, self.p2)
        line_intersection = line.get_intersection_with_line(line2)
        result: list[Point] | list[Segment] = []
        if len(line_intersection) > 0 and isinstance(line_intersection[0], Point):
            sorted_xs = sorted([self.p1.x, self.p2.x])
            sorted_ys = sorted([self.p1.y, self.p2.y])
            if (
                line_intersection[0].x >= sorted_xs[0]
                and line_intersection[0].x <= sorted_xs[-1]
                and line_intersection[0].y >= sorted_ys[0]
                and line_intersection[0].y <= sorted_ys[-1]
            ):
                result = [line_intersection[0]]
        elif len(line_intersection) > 0:
            result = [self]
        return result

    def get_position_at_fraction(self, fraction: float) -> Point:
        """Get point at a fraction along the segment.

        Args:
            fraction: Fraction from 0 to 1 (0 = start, 1 = end).

        Returns:
            The point at that fraction.
        """
        x = self.p1.x + fraction * (self.p2.x - self.p1.x)
        y = self.p1.y + fraction * (self.p2.y - self.p1.y)
        return Point(x, y)

    def get_angle_at_fraction(self, fraction: float) -> float:
        """Get angle at a fraction along the segment.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            Angle in radians.
        """
        return self.get_angle_to_horizontal()

    def get_position_and_angle_at_fraction(
        self, fraction: float
    ) -> tuple[Point, float]:
        """Get both position and angle at a fraction.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            Tuple of (point, angle in radians).
        """
        return (
            self.get_position_at_fraction(fraction),
            self.get_angle_at_fraction(fraction),
        )

    def shortened(
        self,
        length: float,
        start_or_end: typing.Literal["start", "end"] = "end",
    ) -> "Segment":
        """Return a shortened copy of the segment.

        Args:
            length: Amount to shorten by.
            start_or_end: Which end to shorten from.

        Returns:
            A new shortened Segment.
        """
        if length == 0 or self.length() == 0:
            return copy.deepcopy(self)
        if start_or_end == "start":
            return self.reversed().shortened(length).reversed()
        fraction = 1 - length / self.length()
        point = self.get_position_at_fraction(fraction)
        return Segment(self.p1, point)

    def transformed(self, transformation: "Transformation") -> "Segment":
        """Apply a transformation to this segment.

        Args:
            transformation: The transformation to apply.

        Returns:
            A new transformed Segment.
        """
        return Segment(
            self.p1.transformed(transformation),
            self.p2.transformed(transformation),
        )

    def reversed(self) -> "Segment":
        """Return a reversed copy of the segment.

        Returns:
            A new Segment with p1 and p2 swapped.
        """
        return Segment(self.p2, self.p1)

    def bbox(self) -> "Bbox":
        """Get the bounding box of the segment.

        Returns:
            A Bbox enclosing the segment.
        """
        return Bbox.around_points([self.p1, self.p2])


# Precomputed 16-point Gauss-Legendre nodes and weights on [-1, 1]
_GL_NODES_16 = [
    -0.9894009349916499,
    -0.9445750230732326,
    -0.8656312023878318,
    -0.7554044083550030,
    -0.6178762444026438,
    -0.4580167776572274,
    -0.2816035507792589,
    -0.0950125098376374,
    0.0950125098376374,
    0.2816035507792589,
    0.4580167776572274,
    0.6178762444026438,
    0.7554044083550030,
    0.8656312023878318,
    0.9445750230732326,
    0.9894009349916499,
]
_GL_WEIGHTS_16 = [
    0.0271524594117541,
    0.0622535239386479,
    0.0951585116824928,
    0.1246289712555339,
    0.1495959888165767,
    0.1691565193950025,
    0.1826034150449236,
    0.1894506104550685,
    0.1894506104550685,
    0.1826034150449236,
    0.1691565193950025,
    0.1495959888165767,
    0.1246289712555339,
    0.0951585116824928,
    0.0622535239386479,
    0.0271524594117541,
]


def _arc_length(derivative_function, t0: float, t1: float) -> float:
    """Gauss-Legendre quadrature of |derivative(t)| from t0 to t1.

    Args:
        derivative_function: Function returning (dx, dy) at parameter t.
        t0: Start parameter.
        t1: End parameter.

    Returns:
        The arc length between t0 and t1.
    """
    mid = (t0 + t1) / 2
    half = (t1 - t0) / 2
    total = 0.0
    for node, weight in zip(_GL_NODES_16, _GL_WEIGHTS_16):
        t = mid + half * node
        dx, dy = derivative_function(t)
        total += weight * math.sqrt(dx * dx + dy * dy)
    return total * half


def _find_t_at_arc_length_fraction(
    derivative_function,
    total_length: float,
    fraction: float,
    tolerance: float = CONVERGENCE_TOLERANCE,
) -> float:
    """Binary search for t such that arc_length(0, t) / total_length ~ fraction.

    Args:
        derivative_function: Function returning (dx, dy) at parameter t.
        total_length: Total arc length of the curve.
        fraction: Target fraction of total length.
        tolerance: Tolerance (unused, kept for API).

    Returns:
        Parameter t corresponding to the fraction.
    """
    target = fraction * total_length
    low, high = 0.0, 1.0
    for _ in range(64):
        mid = (low + high) / 2
        if _arc_length(derivative_function, 0.0, mid) < target:
            low = mid
        else:
            high = mid
    return (low + high) / 2


@dataclasses.dataclass(frozen=True)
class QuadraticBezierCurve(GeometryObject):
    """Represents a quadratic Bezier curve.

    Attributes:
        p1: Start point.
        p2: End point.
        control_point: The single control point.

    Examples:
        >>> curve = QuadraticBezierCurve(
        ...     Point(0, 0),
        ...     Point(10, 0),
        ...     Point(5, 5)
        ... )
    """

    p1: Point
    p2: Point
    control_point: Point

    def evaluate(self, t: float) -> Point:
        """Evaluate the curve at parameter t.

        Args:
            t: Parameter value from 0 to 1.

        Returns:
            The point at parameter t.
        """
        return self.evaluate_multi([t])[0]

    def evaluate_multi(
        self, t_sequence: collections.abc.Sequence[float]
    ) -> list[Point]:
        """Evaluate the curve at multiple parameters.

        Args:
            t_sequence: Sequence of parameter values.

        Returns:
            List of points.
        """
        t = numpy.asarray(t_sequence, dtype="double")
        u = 1 - t
        x = u * u * self.p1.x + 2 * u * t * self.control_point.x + t * t * self.p2.x
        y = u * u * self.p1.y + 2 * u * t * self.control_point.y + t * t * self.p2.y
        return [Point(float(xi), float(yi)) for xi, yi in zip(x, y)]

    def derivative(self, t: float) -> tuple[float, float]:
        """Compute the derivative at parameter t.

        Args:
            t: Parameter value from 0 to 1.

        Returns:
            Tuple of (dx/dt, dy/dt).
        """
        u = 1 - t
        dx = 2 * u * (self.control_point.x - self.p1.x) + 2 * t * (
            self.p2.x - self.control_point.x
        )
        dy = 2 * u * (self.control_point.y - self.p1.y) + 2 * t * (
            self.p2.y - self.control_point.y
        )
        return (dx, dy)

    def split(self, t: float) -> tuple["QuadraticBezierCurve", "QuadraticBezierCurve"]:
        """Split the curve at parameter t using De Casteljau subdivision.

        Args:
            t: Parameter value from 0 to 1.

        Returns:
            Tuple of two QuadraticBezierCurves.
        """
        u = 1 - t
        m0x = u * self.p1.x + t * self.control_point.x
        m0y = u * self.p1.y + t * self.control_point.y
        m1x = u * self.control_point.x + t * self.p2.x
        m1y = u * self.control_point.y + t * self.p2.y
        mx = u * m0x + t * m1x
        my = u * m0y + t * m1y
        mid = Point(mx, my)
        left = QuadraticBezierCurve(self.p1, mid, Point(m0x, m0y))
        right = QuadraticBezierCurve(mid, self.p2, Point(m1x, m1y))
        return (left, right)

    def length(self) -> float:
        """Calculate the arc length of the curve.

        Returns:
            The arc length.
        """
        return _arc_length(self.derivative, 0.0, 1.0)

    def bbox(self) -> "Bbox":
        """Get the bounding box of the curve.

        Returns:
            A Bbox enclosing the curve.
        """
        # Extrema candidates: endpoints + roots of derivative
        xs = [self.p1.x, self.p2.x]
        ys = [self.p1.y, self.p2.y]
        # dx/dt = 2(1-t)(cp-p1) + 2t(p2-cp) = 0 => t = (p1-cp)/(p1-2cp+p2)
        for values, lst in [
            ((self.p1.x, self.control_point.x, self.p2.x), xs),
            ((self.p1.y, self.control_point.y, self.p2.y), ys),
        ]:
            a, b, c = values
            denominator = a - 2 * b + c
            if abs(denominator) > ZERO_TOLERANCE:
                t = (a - b) / denominator
                if 0 < t < 1:
                    point = self.evaluate(t)
                    lst.append(point.x if lst is xs else point.y)
        min_x, max_x = min(xs), max(xs)
        min_y, max_y = min(ys), max(ys)
        return Bbox(
            Point(min_x, min_y),
            max_x - min_x,
            max_y - min_y,
        )

    def get_position_at_fraction(self, fraction: float) -> Point:
        """Get point at a fraction of the arc length.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            The point at that fraction.
        """
        if fraction <= 0:
            return Point(self.p1.x, self.p1.y)
        if fraction >= 1:
            return Point(self.p2.x, self.p2.y)
        total = self.length()
        t = _find_t_at_arc_length_fraction(self.derivative, total, fraction)
        return self.evaluate(t)

    def get_angle_at_fraction(self, fraction: float) -> float:
        """Get angle at a fraction of the arc length.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            Angle in radians.
        """
        total = self.length()
        if fraction <= 0:
            t = 0.0
        elif fraction >= 1:
            t = 1.0
        else:
            t = _find_t_at_arc_length_fraction(self.derivative, total, fraction)
        dx, dy = self.derivative(t)
        return math.atan2(dy, dx)

    def get_position_and_angle_at_fraction(
        self, fraction: float
    ) -> tuple[Point, float]:
        """Get both position and angle at a fraction of the arc length.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            Tuple of (point, angle_in_radians).
        """
        total = self.length()
        if fraction <= 0:
            t = 0.0
        elif fraction >= 1:
            t = 1.0
        else:
            t = _find_t_at_arc_length_fraction(self.derivative, total, fraction)
        pos = self.evaluate(t)
        dx, dy = self.derivative(t)
        angle = math.atan2(dy, dx)
        return (pos, angle)

    def get_intersection_with_line(self, line: Line) -> list[Point] | list[Segment]:
        """Get intersection with a line.

        Args:
            line: The line to intersect with.

        Returns:
            List of intersection points or segments.
        """
        # Line equation: a*x + b*y + c = 0
        a = line.p2.y - line.p1.y
        b = line.p1.x - line.p2.x
        c = line.p2.x * line.p1.y - line.p1.x * line.p2.y
        # Bezier: B(t) = (1-t)^2 * p0 + 2(1-t)t * p1 + t^2 * p2
        p0x, p0y = self.p1.x, self.p1.y
        p1x, p1y = self.control_point.x, self.control_point.y
        p2x, p2y = self.p2.x, self.p2.y
        # Substitute into line equation: A*t^2 + B*t + C = 0
        ax_coeff = a * (p0x - 2 * p1x + p2x) + b * (p0y - 2 * p1y + p2y)
        bx_coeff = a * (-2 * p0x + 2 * p1x) + b * (-2 * p0y + 2 * p1y)
        cx_coeff = a * p0x + b * p0y + c
        roots = []
        if abs(ax_coeff) < ZERO_TOLERANCE:
            if abs(bx_coeff) > ZERO_TOLERANCE:
                roots.append(-cx_coeff / bx_coeff)
        else:
            disc = bx_coeff * bx_coeff - 4 * ax_coeff * cx_coeff
            if disc >= 0:
                sqrt_disc = math.sqrt(disc)
                roots.append((-bx_coeff + sqrt_disc) / (2 * ax_coeff))
                roots.append((-bx_coeff - sqrt_disc) / (2 * ax_coeff))
        result = []
        for t in roots:
            if -PARAMETER_TOLERANCE <= t <= 1 + PARAMETER_TOLERANCE:
                t = max(0.0, min(1.0, t))
                result.append(self.evaluate(t))
        return result

    def shortened(
        self, length: float, start_or_end: typing.Literal["start", "end"] = "end"
    ) -> "QuadraticBezierCurve":
        """Return a shortened copy of the curve.

        Args:
            length: Amount to shorten by.
            start_or_end: Which end to shorten from.

        Returns:
            A new shortened QuadraticBezierCurve.
        """
        if length == 0 or self.length() == 0:
            return copy.deepcopy(self)
        if start_or_end == "start":
            return self.reversed().shortened(length).reversed()
        total_length = self.length()
        if length > total_length:
            length = total_length
        fraction = 1 - length / total_length
        t = _find_t_at_arc_length_fraction(self.derivative, total_length, fraction)
        left, _ = self.split(t)
        return left

    def reversed(self) -> "QuadraticBezierCurve":
        """Return a reversed copy of the curve.

        Returns:
            A new QuadraticBezierCurve going in reverse direction.
        """
        return QuadraticBezierCurve(self.p2, self.p1, self.control_point)

    def transformed(self, transformation) -> "QuadraticBezierCurve":
        """Apply a transformation to this curve.

        Args:
            transformation: The transformation to apply.

        Returns:
            A new transformed QuadraticBezierCurve.
        """
        return QuadraticBezierCurve(
            self.p1.transformed(transformation),
            self.p2.transformed(transformation),
            self.control_point.transformed(transformation),
        )


@dataclasses.dataclass(frozen=True)
class CubicBezierCurve(GeometryObject):
    """Represents a cubic Bezier curve.

    Attributes:
        p1: Start point.
        p2: End point.
        control_point1: First control point.
        control_point2: Second control point.

    Examples:
        >>> curve = CubicBezierCurve(
        ...     Point(0, 0),
        ...     Point(10, 0),
        ...     Point(3, 5),
        ...     Point(7, 5)
        ... )
    """

    p1: Point
    p2: Point
    control_point1: Point
    control_point2: Point

    def evaluate(self, t: float) -> Point:
        """Evaluate the curve at parameter t.

        Args:
            t: Parameter value from 0 to 1.

        Returns:
            The point at parameter t.
        """
        return self.evaluate_multi([t])[0]

    def evaluate_multi(
        self, t_sequence: collections.abc.Sequence[float]
    ) -> list[Point]:
        """Evaluate the curve at multiple parameters.

        Args:
            t_sequence: Sequence of parameter values.

        Returns:
            List of points.
        """
        t = numpy.asarray(t_sequence, dtype="double")
        u = 1 - t
        u2 = u * u
        t2 = t * t
        x = (
            u2 * u * self.p1.x
            + 3 * u2 * t * self.control_point1.x
            + 3 * u * t2 * self.control_point2.x
            + t2 * t * self.p2.x
        )
        y = (
            u2 * u * self.p1.y
            + 3 * u2 * t * self.control_point1.y
            + 3 * u * t2 * self.control_point2.y
            + t2 * t * self.p2.y
        )
        return [Point(float(xi), float(yi)) for xi, yi in zip(x, y)]

    def derivative(self, t: float) -> tuple[float, float]:
        """Compute the derivative at parameter t.

        Args:
            t: Parameter value from 0 to 1.

        Returns:
            Tuple of (dx/dt, dy/dt).
        """
        u = 1 - t
        dx = (
            3 * u * u * (self.control_point1.x - self.p1.x)
            + 6 * u * t * (self.control_point2.x - self.control_point1.x)
            + 3 * t * t * (self.p2.x - self.control_point2.x)
        )
        dy = (
            3 * u * u * (self.control_point1.y - self.p1.y)
            + 6 * u * t * (self.control_point2.y - self.control_point1.y)
            + 3 * t * t * (self.p2.y - self.control_point2.y)
        )
        return (dx, dy)

    def split(self, t: float) -> tuple["CubicBezierCurve", "CubicBezierCurve"]:
        """Split the curve at parameter t using De Casteljau subdivision.

        Args:
            t: Parameter value from 0 to 1.

        Returns:
            Tuple of two CubicBezierCurves.
        """
        u = 1 - t
        # Level 1
        m0x = u * self.p1.x + t * self.control_point1.x
        m0y = u * self.p1.y + t * self.control_point1.y
        m1x = u * self.control_point1.x + t * self.control_point2.x
        m1y = u * self.control_point1.y + t * self.control_point2.y
        m2x = u * self.control_point2.x + t * self.p2.x
        m2y = u * self.control_point2.y + t * self.p2.y
        # Level 2
        n0x = u * m0x + t * m1x
        n0y = u * m0y + t * m1y
        n1x = u * m1x + t * m2x
        n1y = u * m1y + t * m2y
        # Level 3 (point on curve)
        px = u * n0x + t * n1x
        py = u * n0y + t * n1y
        mid = Point(px, py)
        left = CubicBezierCurve(self.p1, mid, Point(m0x, m0y), Point(n0x, n0y))
        right = CubicBezierCurve(mid, self.p2, Point(n1x, n1y), Point(m2x, m2y))
        return (left, right)

    def length(self) -> float:
        """Calculate the arc length of the curve.

        Returns:
            The arc length.
        """
        return _arc_length(self.derivative, 0.0, 1.0)

    def bbox(self) -> "Bbox":
        """Get the bounding box of the curve.

        Returns:
            A Bbox enclosing the curve.
        """
        xs = [self.p1.x, self.p2.x]
        ys = [self.p1.y, self.p2.y]
        for values, candidates in [
            (
                (self.p1.x, self.control_point1.x, self.control_point2.x, self.p2.x),
                xs,
            ),
            (
                (self.p1.y, self.control_point1.y, self.control_point2.y, self.p2.y),
                ys,
            ),
        ]:
            p0, p1, p2, p3 = values
            a = -3 * p0 + 9 * p1 - 9 * p2 + 3 * p3
            b = 6 * p0 - 12 * p1 + 6 * p2
            c = -3 * p0 + 3 * p1
            if abs(a) < ZERO_TOLERANCE:
                if abs(b) > ZERO_TOLERANCE:
                    t = -c / b
                    if 0 < t < 1:
                        point = self.evaluate(t)
                        candidates.append(point.x if candidates is xs else point.y)
            else:
                disc = b * b - 4 * a * c
                if disc >= 0:
                    sqrt_disc = math.sqrt(disc)
                    for t in [(-b + sqrt_disc) / (2 * a), (-b - sqrt_disc) / (2 * a)]:
                        if 0 < t < 1:
                            point = self.evaluate(t)
                            candidates.append(point.x if candidates is xs else point.y)
        min_x, max_x = min(xs), max(xs)
        min_y, max_y = min(ys), max(ys)
        return Bbox(
            Point(min_x, min_y),
            max_x - min_x,
            max_y - min_y,
        )

    def get_position_at_fraction(self, fraction: float) -> Point:
        """Get point at a fraction of the arc length.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            The point at that fraction.
        """
        if fraction <= 0:
            return Point(self.p1.x, self.p1.y)
        if fraction >= 1:
            return Point(self.p2.x, self.p2.y)
        total = self.length()
        t = _find_t_at_arc_length_fraction(self.derivative, total, fraction)
        return self.evaluate(t)

    def get_angle_at_fraction(self, fraction: float) -> float:
        """Get angle at a fraction of the arc length.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            Angle in radians.
        """
        total = self.length()
        if fraction <= 0:
            t = 0.0
        elif fraction >= 1:
            t = 1.0
        else:
            t = _find_t_at_arc_length_fraction(self.derivative, total, fraction)
        dx, dy = self.derivative(t)
        return math.atan2(dy, dx)

    def get_position_and_angle_at_fraction(
        self, fraction: float
    ) -> tuple[Point, float]:
        """Get both position and angle at a fraction of the arc length.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            Tuple of (point, angle_in_radians).
        """
        total = self.length()
        if fraction <= 0:
            t = 0.0
        elif fraction >= 1:
            t = 1.0
        else:
            t = _find_t_at_arc_length_fraction(self.derivative, total, fraction)
        pos = self.evaluate(t)
        dx, dy = self.derivative(t)
        angle = math.atan2(dy, dx)
        return (pos, angle)

    def get_intersection_with_line(self, line: Line) -> list[Point] | list[Segment]:
        """Get intersection with a line.

        Args:
            line: The line to intersect with.

        Returns:
            List of intersection points or segments.
        """
        a = line.p2.y - line.p1.y
        b = line.p1.x - line.p2.x
        c = line.p2.x * line.p1.y - line.p1.x * line.p2.y
        p0x, p0y = self.p1.x, self.p1.y
        p1x, p1y = self.control_point1.x, self.control_point1.y
        p2x, p2y = self.control_point2.x, self.control_point2.y
        p3x, p3y = self.p2.x, self.p2.y
        c3x = -p0x + 3 * p1x - 3 * p2x + p3x
        c2x = 3 * p0x - 6 * p1x + 3 * p2x
        c1x = -3 * p0x + 3 * p1x
        c0x = p0x
        c3y = -p0y + 3 * p1y - 3 * p2y + p3y
        c2y = 3 * p0y - 6 * p1y + 3 * p2y
        c1y = -3 * p0y + 3 * p1y
        c0y = p0y
        coeff3 = a * c3x + b * c3y
        coeff2 = a * c2x + b * c2y
        coeff1 = a * c1x + b * c1y
        coeff0 = a * c0x + b * c0y + c
        roots = numpy.roots([coeff3, coeff2, coeff1, coeff0])
        result = []
        for root in roots:
            if abs(root.imag) < CONVERGENCE_TOLERANCE:
                t = root.real
                if -PARAMETER_TOLERANCE <= t <= 1 + PARAMETER_TOLERANCE:
                    t = max(0.0, min(1.0, t))
                    result.append(self.evaluate(t))
        return result

    def shortened(
        self, length: float, start_or_end: typing.Literal["start", "end"] = "end"
    ) -> "CubicBezierCurve":
        """Return a shortened copy of the curve.

        Args:
            length: Amount to shorten by.
            start_or_end: Which end to shorten from.

        Returns:
            A new shortened CubicBezierCurve.
        """
        if length == 0 or self.length() == 0:
            return copy.deepcopy(self)
        if start_or_end == "start":
            return self.reversed().shortened(length).reversed()
        total_length = self.length()
        if length > total_length:
            length = total_length
        fraction = 1 - length / total_length
        t = _find_t_at_arc_length_fraction(self.derivative, total_length, fraction)
        left, _ = self.split(t)
        return left

    def reversed(self) -> "CubicBezierCurve":
        """Return a reversed copy of the curve.

        Returns:
            A new CubicBezierCurve going in reverse direction.
        """
        return CubicBezierCurve(
            self.p2, self.p1, self.control_point2, self.control_point1
        )

    def transformed(self, transformation) -> "CubicBezierCurve":
        """Apply a transformation to this curve.

        Args:
            transformation: The transformation to apply.

        Returns:
            A new transformed CubicBezierCurve.
        """
        return CubicBezierCurve(
            self.p1.transformed(transformation),
            self.p2.transformed(transformation),
            self.control_point1.transformed(transformation),
            self.control_point2.transformed(transformation),
        )


@dataclasses.dataclass(frozen=True)
class EllipticalArc(GeometryObject):
    """Represents an elliptical arc.

    Attributes:
        p1: Start point.
        p2: End point.
        rx: X-radius of the ellipse.
        ry: Y-radius of the ellipse.
        x_axis_rotation: Rotation of the x-axis in radians.
        arc_flag: Large arc flag (0 or 1).
        sweep_flag: Sweep flag (0 or 1).

    Examples:
        >>> arc = EllipticalArc(
        ...     Point(0, 0), Point(10, 0), 5, 5, 0, 0, 1
        ... )
    """

    p1: Point
    p2: Point
    rx: float
    ry: float
    x_axis_rotation: float
    arc_flag: int
    sweep_flag: int

    def __post_init__(self):
        object.__setattr__(self, "rx", round(self.rx, ROUNDING))
        object.__setattr__(self, "ry", round(self.ry, ROUNDING))

    def get_intersection_with_line(self, line: Line) -> list[Point]:
        """Get intersection with a line.

        Uses analytical ellipse-line intersection by solving
        A*cos(theta) + B*sin(theta) + C = 0 via atan2.

        Args:
            line: The line to intersect with.

        Returns:
            List of intersection points.
        """
        cx, cy, rx, ry, sigma, theta1, theta2, delta_theta = (
            self.get_center_parameterization()
        )
        # Line in implicit form: a*x + b*y + c = 0
        dx = line.p2.x - line.p1.x
        dy = line.p2.y - line.p1.y
        a = -dy
        b = dx
        c = -(a * line.p1.x + b * line.p1.y)
        cos_sigma = math.cos(sigma)
        sin_sigma = math.sin(sigma)
        # Substitute parametric arc equations into line equation
        A = a * rx * cos_sigma + b * rx * sin_sigma
        B = a * (-ry * sin_sigma) + b * ry * cos_sigma
        C = a * cx + b * cy + c
        R = math.sqrt(A * A + B * B)
        if R < ZERO_TOLERANCE:
            return []
        ratio = -C / R
        if abs(ratio) > 1 + ZERO_TOLERANCE:
            return []
        ratio = max(-1.0, min(1.0, ratio))
        phi = math.atan2(B, A)
        acos_val = math.acos(ratio)
        candidates = [phi + acos_val, phi - acos_val]
        if delta_theta > 0:
            theta_min = theta1
            theta_max = theta1 + delta_theta
        else:
            theta_min = theta1 + delta_theta
            theta_max = theta1
        result = []
        for theta_candidate in candidates:
            # Normalize into range by trying multiple periods
            for k in range(-3, 4):
                theta_c = theta_candidate + k * 2 * math.pi
                if theta_min - ZERO_TOLERANCE <= theta_c <= theta_max + ZERO_TOLERANCE:
                    t = (theta_c - theta1) / delta_theta
                    if -ZERO_TOLERANCE <= t <= 1 + ZERO_TOLERANCE:
                        t = max(0.0, min(1.0, t))
                        point = self.evaluate(t)
                        # Avoid duplicates
                        is_dup = False
                        for existing in result:
                            if (
                                abs(existing.x - point.x) < ROUNDING_TOLERANCE
                                and abs(existing.y - point.y)
                                < ROUNDING_TOLERANCE
                            ):
                                is_dup = True
                                break
                        if not is_dup:
                            result.append(point)
        return result

    def get_center_parameterization(
        self,
    ) -> tuple[float, float, float, float, float, float, float, float]:
        """Get the center parameterization of the arc.

        Returns:
            Tuple of (cx, cy, rx, ry, sigma, theta1, theta2, delta_theta).
        """
        x1, y1 = self.p1.x, self.p1.y
        sigma = self.x_axis_rotation
        x2, y2 = self.p2.x, self.p2.y
        rx = self.rx
        ry = self.ry
        fa = self.arc_flag
        fs = self.sweep_flag
        x1p = math.cos(sigma) * ((x1 - x2) / 2) + math.sin(sigma) * ((y1 - y2) / 2)
        y1p = -math.sin(sigma) * ((x1 - x2) / 2) + math.cos(sigma) * ((y1 - y2) / 2)
        l = x1p**2 / rx**2 + y1p**2 / ry**2
        if l > 1:
            rx = math.sqrt(l) * rx
            ry = math.sqrt(l) * ry
        r = rx**2 * ry**2 - rx**2 * y1p**2 - ry**2 * x1p**2
        if r < 0:
            r = 0
        a = math.sqrt(r / (rx**2 * y1p**2 + ry**2 * x1p**2))
        if fa == fs:
            a = -a
        cxp = a * rx * y1p / ry
        cyp = -a * ry * x1p / rx
        cx = math.cos(sigma) * cxp - math.sin(sigma) * cyp + (x1 + x2) / 2
        cy = math.sin(sigma) * cxp + math.cos(sigma) * cyp + (y1 + y2) / 2
        theta1 = _get_angle_between_segments(
            Segment(Point(0, 0), Point(1, 0)),
            Segment(Point(0, 0), Point((x1p - cxp) / rx, (y1p - cyp) / ry)),
        )
        delta_theta = _get_angle_between_segments(
            Segment(Point(0, 0), Point((x1p - cxp) / rx, (y1p - cyp) / ry)),
            Segment(Point(0, 0), Point(-(x1p + cxp) / rx, -(y1p + cyp) / ry)),
        )
        if fs == 0 and delta_theta > 0:
            delta_theta -= 2 * math.pi
        elif fs == 1 and delta_theta < 0:
            delta_theta += 2 * math.pi
        theta2 = theta1 + delta_theta
        return cx, cy, rx, ry, sigma, theta1, theta2, delta_theta

    def get_center(self) -> Point:
        """Get the center point of the ellipse.

        Returns:
            The center point.
        """
        cx, cy, *_ = self.get_center_parameterization()
        return Point(cx, cy)

    def evaluate(self, t: float) -> Point:
        """Evaluate the arc at parameter t using center parameterization.

        Args:
            t: Parameter value from 0 to 1.

        Returns:
            The point at parameter t.
        """
        cx, cy, rx, ry, sigma, theta1, theta2, delta_theta = (
            self.get_center_parameterization()
        )
        theta = theta1 + t * delta_theta
        cos_sigma = math.cos(sigma)
        sin_sigma = math.sin(sigma)
        cos_theta = math.cos(theta)
        sin_theta = math.sin(theta)
        x = cx + rx * cos_theta * cos_sigma - ry * sin_theta * sin_sigma
        y = cy + rx * cos_theta * sin_sigma + ry * sin_theta * cos_sigma
        return Point(x, y)

    def derivative(self, t: float) -> tuple[float, float]:
        """Compute the derivative of the arc at parameter t.

        Args:
            t: Parameter value from 0 to 1.

        Returns:
            Tuple of (dx, dy) derivatives.
        """
        cx, cy, rx, ry, sigma, theta1, theta2, delta_theta = (
            self.get_center_parameterization()
        )
        theta = theta1 + t * delta_theta
        cos_sigma = math.cos(sigma)
        sin_sigma = math.sin(sigma)
        cos_theta = math.cos(theta)
        sin_theta = math.sin(theta)
        dx = delta_theta * (
            -rx * sin_theta * cos_sigma - ry * cos_theta * sin_sigma
        )
        dy = delta_theta * (
            -rx * sin_theta * sin_sigma + ry * cos_theta * cos_sigma
        )
        return (dx, dy)

    def get_position_at_fraction(self, fraction: float) -> Point:
        """Get point at a fraction along the arc.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            The point at that fraction.
        """
        if fraction <= 0:
            return Point(self.p1.x, self.p1.y)
        if fraction >= 1:
            return Point(self.p2.x, self.p2.y)
        total = self.length()
        if total == 0:
            return Point(self.p1.x, self.p1.y)
        t = _find_t_at_arc_length_fraction(self.derivative, total, fraction)
        return self.evaluate(t)

    def get_angle_at_fraction(self, fraction: float) -> float:
        """Get angle at a fraction along the arc.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            Angle in radians.
        """
        total = self.length()
        if total == 0:
            return 0.0
        if fraction <= 0:
            t = 0.0
        elif fraction >= 1:
            t = 1.0
        else:
            t = _find_t_at_arc_length_fraction(self.derivative, total, fraction)
        dx, dy = self.derivative(t)
        return math.atan2(dy, dx)

    def get_position_and_angle_at_fraction(
        self, fraction: float
    ) -> tuple[Point, float]:
        """Get both position and angle at a fraction.

        Args:
            fraction: Fraction from 0 to 1.

        Returns:
            Tuple of (point, angle).
        """
        total = self.length()
        if total == 0:
            return (Point(self.p1.x, self.p1.y), 0.0)
        if fraction <= 0:
            t = 0.0
        elif fraction >= 1:
            t = 1.0
        else:
            t = _find_t_at_arc_length_fraction(
                self.derivative, total, fraction
            )
        pos = self.evaluate(t)
        dx, dy = self.derivative(t)
        angle = math.atan2(dy, dx)
        return (pos, angle)

    def bbox(self) -> "Bbox":
        """Get the bounding box of the arc.

        Returns:
            A Bbox enclosing the arc.
        """
        cx, cy, rx, ry, sigma, theta1, theta2, delta_theta = (
            self.get_center_parameterization()
        )
        cos_sigma = math.cos(sigma)
        sin_sigma = math.sin(sigma)
        # Extrema of x(theta): dx/dtheta = 0
        # -rx*sin(theta)*cos(sigma) - ry*cos(theta)*sin(sigma) = 0
        # tan(theta) = -(ry*sin(sigma)) / (rx*cos(sigma))
        theta_x = math.atan2(-ry * sin_sigma, rx * cos_sigma)
        # Extrema of y(theta): dy/dtheta = 0
        # -rx*sin(theta)*sin(sigma) + ry*cos(theta)*cos(sigma) = 0
        # tan(theta) = (ry*cos(sigma)) / (rx*sin(sigma))
        theta_y = math.atan2(ry * cos_sigma, rx * sin_sigma)
        # Collect candidate points: endpoints + extrema within arc range
        points = [self.p1, self.p2]
        if delta_theta > 0:
            theta_min = theta1
            theta_max = theta1 + delta_theta
        else:
            theta_min = theta1 + delta_theta
            theta_max = theta1
        for theta_base in (theta_x, theta_y):
            # Check both solutions (theta_base and theta_base + pi)
            for candidate in (theta_base, theta_base + math.pi):
                # Normalize candidate into [theta_min - 2pi, theta_max + 2pi]
                # and check if it falls within range
                normalized = candidate
                while normalized < theta_min - 2 * math.pi:
                    normalized += 2 * math.pi
                while normalized > theta_max + 2 * math.pi:
                    normalized -= 2 * math.pi
                # Try multiple periods
                for offset in (
                    -2 * math.pi,
                    0,
                    2 * math.pi,
                ):
                    theta_c = normalized + offset
                    if theta_min - ZERO_TOLERANCE <= theta_c <= theta_max + ZERO_TOLERANCE:
                        t = (theta_c - theta1) / delta_theta
                        if -ZERO_TOLERANCE <= t <= 1 + ZERO_TOLERANCE:
                            points.append(self.evaluate(max(0, min(1, t))))
        return Bbox.around_points(points)

    def shortened(
        self,
        length: float,
        start_or_end: typing.Literal["start", "end"] = "end",
    ) -> "EllipticalArc":
        """Return a shortened copy of the arc.

        Args:
            length: Amount to shorten by.
            start_or_end: Which end to shorten from.

        Returns:
            A new shortened EllipticalArc.
        """
        if length == 0 or self.length() == 0:
            return copy.deepcopy(self)
        if start_or_end == "start":
            return self.reversed().shortened(length).reversed()
        fraction = 1 - length / self.length()
        point = self.get_position_at_fraction(fraction)
        return dataclasses.replace(self, p2=point)

    def transformed(self, transformation: "Transformation") -> "EllipticalArc":
        """Apply a transformation to this arc.

        Args:
            transformation: The transformation to apply.

        Returns:
            A new transformed EllipticalArc.
        """
        east = Point(
            math.cos(self.x_axis_rotation) * self.rx,
            math.sin(self.x_axis_rotation) * self.rx,
        )
        north = Point(
            math.cos(self.x_axis_rotation) * self.ry,
            math.sin(self.x_axis_rotation) * self.ry,
        )
        new_center = Point(0, 0).transformed(transformation)
        new_east = east.transformed(transformation)
        new_north = north.transformed(transformation)
        new_rx = Segment(new_center, new_east).length()
        new_ry = Segment(new_center, new_north).length()
        new_start_point = self.p1.transformed(transformation)
        new_end_point = self.p2.transformed(transformation)
        new_x_axis_rotation = math.degrees(
            Line(new_center, new_east).get_angle_to_horizontal()
        )
        return EllipticalArc(
            p1=new_start_point,
            p2=new_end_point,
            rx=new_rx,
            ry=new_ry,
            x_axis_rotation=new_x_axis_rotation,
            arc_flag=self.arc_flag,
            sweep_flag=self.sweep_flag,
        )

    def reversed(self) -> "EllipticalArc":
        """Return a reversed copy of the arc.

        Returns:
            A new EllipticalArc going in reverse direction.
        """
        return EllipticalArc(
            self.p2,
            self.p1,
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.arc_flag,
            abs(self.sweep_flag - 1),
        )

    def length(self) -> float:
        """Calculate the length of the arc.

        Returns:
            The arc length.
        """
        return _arc_length(self.derivative, 0.0, 1.0)


@dataclasses.dataclass(frozen=True)
class Bbox(object):
    """Represents a bounding box.

    Attributes:
        position: Center point.
        width: Width of the box.
        height: Height of the box.

    Examples:
        >>> bbox = Bbox(Point(5, 5), 10, 10)
        >>> bbox.north_west()
        Point(x=0.0, y=0.0)
        >>> bbox.south_east()
        Point(x=10.0, y=10.0)
    """

    position: Point
    width: float
    height: float

    def __post_init__(self):
        object.__setattr__(self, "width", round(self.width, ROUNDING))
        object.__setattr__(self, "height", round(self.height, ROUNDING))

    @property
    def x(self) -> float:
        """The x coordinate of the center."""
        return self.position.x

    @property
    def y(self) -> float:
        """The y coordinate of the center."""
        return self.position.y

    def size(self) -> tuple[float, float]:
        """Get the size as (width, height).

        Returns:
            Tuple of (width, height).
        """
        return (self.width, self.height)

    def anchor_point(self, anchor_point: str) -> Point:
        """Get a named anchor point.

        Args:
            anchor_point: Name like 'north', 'south_east', 'center', etc.

        Returns:
            The anchor point.
        """
        return getattr(self, anchor_point)()

    def north_west(self) -> Point:
        """Get the north-west corner."""
        return Point(self.x - self.width / 2, self.y - self.height / 2)

    def north_north_west(self) -> Point:
        """Get the north-north-west point."""
        return Point(self.x - self.width / 4, self.y - self.height / 2)

    def north(self) -> Point:
        """Get the north (top center) point."""
        return Point(self.x, self.y - self.height / 2)

    def north_north_east(self) -> Point:
        """Get the north-north-east point."""
        return Point(self.x + self.width / 4, self.y - self.height / 2)

    def north_east(self) -> Point:
        """Get the north-east corner."""
        return Point(self.x + self.width / 2, self.y - self.height / 2)

    def east_north_east(self) -> Point:
        """Get the east-north-east point."""
        return Point(self.x + self.width / 2, self.y - self.height / 4)

    def east(self) -> Point:
        """Get the east (right center) point."""
        return Point(self.x + self.width / 2, self.y)

    def east_south_east(self) -> Point:
        """Get the east-south-east point."""
        return Point(self.x + self.width / 2, self.y + self.width / 4)

    def south_east(self) -> Point:
        """Get the south-east corner."""
        return Point(self.x + self.width / 2, self.y + self.height / 2)

    def south_south_east(self) -> Point:
        """Get the south-south-east point."""
        return Point(self.x + self.width / 4, self.y + self.height / 2)

    def south(self) -> Point:
        """Get the south (bottom center) point."""
        return Point(self.x, self.y + self.height / 2)

    def south_south_west(self) -> Point:
        """Get the south-south-west point."""
        return Point(self.x - self.width / 4, self.y + self.height / 2)

    def south_west(self) -> Point:
        """Get the south-west corner."""
        return Point(self.x - self.width / 2, self.y + self.height / 2)

    def west_south_west(self) -> Point:
        """Get the west-south-west point."""
        return Point(self.x - self.width / 2, self.y + self.height / 4)

    def west(self) -> Point:
        """Get the west (left center) point."""
        return Point(self.x - self.width / 2, self.y)

    def west_north_west(self) -> Point:
        """Get the west-north-west point."""
        return Point(self.x - self.width / 2, self.y - self.height / 4)

    def center(self) -> Point:
        """Get the center point."""
        return Point(self.x, self.y)

    def isnan(self) -> bool:
        """Check if the position has NaN coordinates.

        Returns:
            True if position has NaN.
        """
        return self.position.isnan()

    @classmethod
    def around_points(cls, points: collections.abc.Iterable[Point]) -> "Bbox":
        """Create a minimal Bbox enclosing all given points.

        Args:
            points: An iterable of Point objects.

        Returns:
            A new Bbox that tightly encloses all input points.

        Raises:
            ValueError: If the iterable is empty.
        """
        it = iter(points)
        try:
            first = next(it)
        except StopIteration:
            raise ValueError("points must contain at least one point")
        min_x = max_x = first.x
        min_y = max_y = first.y
        for p in it:
            if p.x < min_x:
                min_x = p.x
            elif p.x > max_x:
                max_x = p.x
            if p.y < min_y:
                min_y = p.y
            elif p.y > max_y:
                max_y = p.y
        return cls(
            Point((min_x + max_x) / 2, (min_y + max_y) / 2),
            max_x - min_x,
            max_y - min_y,
        )

    @classmethod
    def union(cls, bboxes: list["Bbox"]) -> "Bbox":
        """Create a Bbox enclosing all given bounding boxes.

        Args:
            bboxes: List of Bbox objects to merge.

        Returns:
            A new Bbox enclosing all input bboxes.
        """
        if not bboxes:
            return cls(Point(0, 0), 0, 0)
        corners = []
        for bbox in bboxes:
            corners.append(bbox.north_west())
            corners.append(bbox.south_east())
        return cls.around_points(corners)


@dataclasses.dataclass(frozen=True)
class Transformation(abc.ABC):
    """Abstract base class for geometric transformations."""

    @abc.abstractmethod
    def to_matrix(self) -> numpy.typing.NDArray:
        """Convert to a 3x3 transformation matrix.

        Returns:
            A 3x3 numpy array.
        """
        pass

    @abc.abstractmethod
    def inverted(self) -> "Transformation":
        """Get the inverse transformation.

        Returns:
            The inverse transformation.
        """
        pass

    def __mul__(self, other):
        return MatrixTransformation(numpy.matmul(self.to_matrix(), other.to_matrix()))


@dataclasses.dataclass(frozen=True)
class MatrixTransformation(Transformation):
    """Represents a transformation as a 3x3 matrix.

    Attributes:
        m: The 3x3 transformation matrix.
    """

    m: numpy.typing.NDArray

    def to_matrix(self) -> numpy.typing.NDArray:
        """Get the matrix representation.

        Returns:
            The 3x3 matrix.
        """
        return self.m

    def inverted(self) -> Transformation:
        """Get the inverse transformation.

        Returns:
            The inverse matrix transformation.
        """
        return MatrixTransformation(numpy.linalg.inv(self.m))


@dataclasses.dataclass(frozen=True)
class Rotation(Transformation):
    """Represents a rotation transformation.

    Attributes:
        angle: Rotation angle in radians.
        point: Optional center of rotation (defaults to origin).

    Examples:
        >>> rot = Rotation(math.pi / 2, Point(5, 5))
    """

    angle: float
    point: Point | None = None

    def to_matrix(self) -> numpy.typing.NDArray:
        """Convert to a rotation matrix.

        Returns:
            A 3x3 rotation matrix.
        """
        m = numpy.array(
            [
                [math.cos(self.angle), -math.sin(self.angle), 0],
                [math.sin(self.angle), math.cos(self.angle), 0],
                [0, 0, 1],
            ],
            dtype=float,
        )
        if self.point is not None:
            translation = Translation(self.point.x, self.point.y)
            m = numpy.matmul(
                numpy.matmul(translation.to_matrix(), m),
                translation.inverted().to_matrix(),
            )
        return m

    def inverted(self) -> Transformation:
        """Get the inverse rotation.

        Returns:
            A rotation by the negative angle.
        """
        return Rotation(-self.angle, self.point)


@dataclasses.dataclass(frozen=True)
class Translation(Transformation):
    """Represents a translation transformation.

    Attributes:
        tx: Translation in x direction.
        ty: Translation in y direction.

    Examples:
        >>> trans = Translation(10, 20)
    """

    tx: float
    ty: float

    def to_matrix(self) -> numpy.typing.NDArray:
        """Convert to a translation matrix.

        Returns:
            A 3x3 translation matrix.
        """
        m = numpy.array([[1, 0, self.tx], [0, 1, self.ty], [0, 0, 1]], dtype=float)
        return m

    def inverted(self) -> Transformation:
        """Get the inverse translation.

        Returns:
            A translation by (-tx, -ty).
        """
        return Translation(-self.tx, -self.ty)


@dataclasses.dataclass(frozen=True)
class Scaling(Transformation):
    """Represents a scaling transformation.

    Attributes:
        sx: Scale factor in x direction.
        sy: Scale factor in y direction.

    Examples:
        >>> scale = Scaling(2, 2)  # Double size
    """

    sx: float
    sy: float

    def to_matrix(self) -> numpy.typing.NDArray:
        """Convert to a scaling matrix.

        Returns:
            A 3x3 scaling matrix.
        """
        m = numpy.array(
            [
                [self.sx, 0, 0],
                [0, self.sy, 0],
                [0, 0, 1],
            ],
            dtype=float,
        )
        return m

    def inverted(self) -> Transformation:
        """Get the inverse scaling.

        Returns:
            Scaling by (1/sx, 1/sy).
        """
        return Scaling(1 / self.sx, 1 / self.sy)


def _intersect_line_with_primitive(
    line: Line,
    primitive: "Segment | QuadraticBezierCurve | CubicBezierCurve | EllipticalArc",
) -> list[Point]:
    """Get intersection of a line with a geometry primitive.

    Args:
        line: The line.
        primitive: A geometry primitive.

    Returns:
        List of intersection points.
    """
    result = primitive.get_intersection_with_line(line)
    points = []
    for obj in result:
        if isinstance(obj, Point):
            points.append(obj)
        elif isinstance(obj, Segment):
            points.append(obj.p1)
            points.append(obj.p2)
    return points


def get_primitives_border(
    primitives: list[Segment | QuadraticBezierCurve | CubicBezierCurve | EllipticalArc],
    point: Point,
    center: Point | None = None,
) -> Point | None:
    """Get the border point of geometry primitives in a given direction.

    Args:
        primitives: List of geometry primitives.
        point: Direction point.
        center: Optional center point. Defaults to bbox center.

    Returns:
        The border point or None.
    """
    if not primitives:
        return None
    if center is None:
        bboxes = [p.bbox() for p in primitives]
        bbox = Bbox.union(bboxes)
        center = bbox.center()
    if center.isnan():
        return None
    line = Line(center, point)
    candidate_points = []
    for primitive in primitives:
        candidate_points.extend(_intersect_line_with_primitive(line, primitive))
    intersection_point = None
    max_d = -1
    ok_direction_exists = False
    d1 = Segment(point, center).length()
    for candidate_point in candidate_points:
        d2 = Segment(candidate_point, point).length()
        d3 = Segment(candidate_point, center).length()
        candidate_ok_direction = not d2 > d1 or d2 < d3
        if candidate_ok_direction or not ok_direction_exists:
            if candidate_ok_direction and not ok_direction_exists:
                ok_direction_exists = True
                max_d = -1
            if d3 > max_d:
                max_d = d3
                intersection_point = candidate_point
    return intersection_point


def get_primitives_angle(
    primitives: list[Segment | QuadraticBezierCurve | CubicBezierCurve | EllipticalArc],
    angle: float,
    unit: typing.Literal["degrees", "radians"] = "degrees",
    center: Point | None = None,
) -> Point | None:
    """Get the border point at a given angle.

    Args:
        primitives: List of geometry primitives.
        angle: The angle.
        unit: Unit of angle ('degrees' or 'radians').
        center: Optional center point.

    Returns:
        The border point or None.
    """
    if unit == "degrees":
        angle = math.radians(angle)
    angle = -angle
    d = 100
    if center is None:
        bboxes = [p.bbox() for p in primitives]
        bbox = Bbox.union(bboxes)
        center = bbox.center()
        if center.isnan():
            return None
    point = center + (d * math.cos(angle), d * math.sin(angle))
    return get_primitives_border(primitives, point, center)


def get_primitives_anchor_point(
    primitives: list[
        "Segment | QuadraticBezierCurve | CubicBezierCurve | EllipticalArc"
    ],
    anchor_point: str,
    center: Point | None = None,
) -> Point | None:
    """Get an anchor point of geometry primitives.

    Args:
        primitives: List of geometry primitives.
        anchor_point: Name of anchor point.
        center: Optional center point.

    Returns:
        The anchor point or None.
    """
    bboxes = [p.bbox() for p in primitives]
    bbox = Bbox.union(bboxes)
    if center is None:
        center = bbox.center()
    if center.isnan():
        return None
    point = bbox.anchor_point(anchor_point)
    return get_primitives_border(primitives, point, center)


def get_normalized_angle(angle: float) -> float:
    """Normalize an angle to [0, 2*pi).

    Args:
        angle: Angle in radians.

    Returns:
        Normalized angle.
    """
    return angle - (angle // (2 * math.pi) * (2 * math.pi))


def _get_angle_between_segments(segment1: Segment, segment2: Segment) -> float:
    """Get angle between two segments.

    Args:
        segment1: First segment.
        segment2: Second segment.

    Returns:
        Angle in radians (range -pi to pi).
    """
    p1 = segment1.p2 - segment1.p1
    p2 = segment2.p2 - segment2.p1
    scalar_prod = p1.x * p2.x + p1.y * p2.y
    angle = math.acos(
        round(scalar_prod / (segment1.length() * segment2.length()), ROUNDING)
    )
    sign = p1.x * p2.y - p1.y * p2.x
    if sign < 0:
        angle = -angle
    return angle


def get_transformation_for_frame(
    origin: Point, unit_x: Point, unit_y: Point
) -> MatrixTransformation:
    """Get transformation for a frame defined by origin and axes.

    Given a frame F defined by its origin, unit x axis vector, and unit y axis
    vector, returns the transformation that converts points from F coordinates
    to reference frame coordinates.

    Args:
        origin: Origin of the frame.
        unit_x: Unit x-axis vector.
        unit_y: Unit y-axis vector.

    Returns:
        The transformation matrix.
    """
    m = numpy.array(
        [
            [
                unit_x.x - origin.x,
                unit_y.x - origin.x,
                origin.x,
            ],
            [
                unit_x.y - origin.y,
                unit_y.y - origin.y,
                origin.y,
            ],
            [0, 0, 1],
        ],
        dtype=float,
    )
    return MatrixTransformation(m)


def _point_builder_add(self, xy):
    if momapy.builder.isinstance_or_builder(xy, Point):
        xy = (xy.x, xy.y)
    return PointBuilder(self.x + xy[0], self.y + xy[1])


def _point_builder_sub(self, xy):
    if momapy.builder.isinstance_or_builder(xy, Point):
        xy = (xy.x, xy.y)
    return PointBuilder(self.x - xy[0], self.y - xy[1])


def _point_builder_mul(self, scalar):
    return PointBuilder(self.x * scalar, self.y * scalar)


def _point_builder_div(self, scalar):
    return PointBuilder(self.x / scalar, self.y / scalar)


def _point_builder_iter(self):
    yield self.x
    yield self.y


def _point_builder_post_init(self):
    object.__setattr__(self, "x", round(self.x, ROUNDING))
    object.__setattr__(self, "y", round(self.y, ROUNDING))


PointBuilder = momapy.builder.get_or_make_builder_cls(
    Point,
    builder_namespace={
        "__add__": _point_builder_add,
        "__sub__": _point_builder_sub,
        "__mul__": _point_builder_mul,
        "__div__": _point_builder_div,
        "__iter__": _point_builder_iter,
        "__post_init__": _point_builder_post_init,
    },
)

momapy.builder.register_builder_cls(PointBuilder)

BboxBuilder = momapy.builder.get_or_make_builder_cls(Bbox)

momapy.builder.register_builder_cls(BboxBuilder)



"""Font file lookup using uharfbuzz for metadata reading."""

import os
import platform
import dataclasses

import uharfbuzz

import momapy.drawing


@dataclasses.dataclass(frozen=True, kw_only=True)
class _FontEntry:
    """A cached font entry with metadata."""

    path: str
    family: str
    weight: int
    is_italic: bool


_FONT_WEIGHT_MAP: dict[momapy.drawing.FontWeight, int] = {
    momapy.drawing.FontWeight.NORMAL: 400,
    momapy.drawing.FontWeight.BOLD: 700,
    momapy.drawing.FontWeight.BOLDER: 800,
    momapy.drawing.FontWeight.LIGHTER: 300,
}

_FONT_EXTENSIONS = frozenset({".ttf", ".otf", ".ttc"})

_font_cache: list[_FontEntry] | None = None
_query_cache: dict[
    tuple[str, momapy.drawing.FontWeight | int, momapy.drawing.FontStyle],
    str | None,
] = {}


def _get_font_directories() -> list[str]:
    """Return existing system font directories for the current platform.

    Returns:
        List of existing font directory paths.
    """
    system = platform.system()
    home = os.path.expanduser("~")
    if system == "Linux":
        candidates = [
            "/usr/share/fonts",
            "/usr/local/share/fonts",
            os.path.join(home, ".local", "share", "fonts"),
            os.path.join(home, ".fonts"),
        ]
    elif system == "Darwin":
        candidates = [
            "/System/Library/Fonts",
            "/Library/Fonts",
            os.path.join(home, "Library", "Fonts"),
        ]
    elif system == "Windows":
        windir = os.environ.get("WINDIR", r"C:\Windows")
        candidates = [
            os.path.join(windir, "Fonts"),
            os.path.join(home, "AppData", "Local", "Microsoft", "Windows", "Fonts"),
        ]
    else:
        candidates = []
    return [candidate for candidate in candidates if os.path.isdir(candidate)]


def _read_font_metadata(path: str) -> list[_FontEntry]:
    """Read font metadata from a font file using uharfbuzz.

    Args:
        path: Path to the font file.

    Returns:
        List of font entries (multiple for .ttc files).
    """
    entries: list[_FontEntry] = []
    try:
        blob = uharfbuzz.Blob.from_file_path(path)
        face = uharfbuzz.Face(blob)
        num_faces = face.count
    except Exception:
        return entries
    for index in range(num_faces):
        try:
            face = uharfbuzz.Face(blob, index)
            family = face.get_name(1)
            if family is None:
                continue
            family = family.lower()
            subfamily = face.get_name(2)
            subfamily = subfamily.lower() if subfamily else ""
            # Read OS/2 table for weight and italic
            os2_data = face.table_tags
            weight = 400
            is_italic = False
            if b"OS/2" in os2_data:
                os2_blob = face.reference_table(b"OS/2")
                os2_bytes = memoryview(os2_blob)
                if len(os2_bytes) >= 6:
                    weight = int.from_bytes(os2_bytes[4:6], byteorder="big")
                if len(os2_bytes) >= 64:
                    fs_selection = int.from_bytes(os2_bytes[62:64], byteorder="big")
                    is_italic = bool(fs_selection & 1)
            # Fallback: check subfamily string
            if not is_italic and ("italic" in subfamily or "oblique" in subfamily):
                is_italic = True
            entries.append(
                _FontEntry(
                    path=path,
                    family=family,
                    weight=weight,
                    is_italic=is_italic,
                )
            )
        except Exception:
            continue
    return entries


def _scan_fonts() -> list[_FontEntry]:
    """Scan system font directories and return all font entries.

    Returns:
        List of all discovered font entries.
    """
    entries: list[_FontEntry] = []
    for font_dir in _get_font_directories():
        for dirpath, _, filenames in os.walk(font_dir):
            for filename in filenames:
                ext = os.path.splitext(filename)[1].lower()
                if ext in _FONT_EXTENSIONS:
                    path = os.path.join(dirpath, filename)
                    entries.extend(_read_font_metadata(path))
    return entries


def _get_or_make_font_cache() -> list[_FontEntry]:
    """Return the font cache, populating it on first call.

    Returns:
        List of cached font entries.
    """
    global _font_cache, _query_cache
    if _font_cache is None:
        _font_cache = _scan_fonts()
        _query_cache = {}
    return _font_cache


def _score_font(
    entry: _FontEntry,
    family: str,
    weight: int,
    want_italic: bool,
) -> float:
    """Score a font entry against desired properties.

    Lower is better.

    Args:
        entry: The font entry to score.
        family: Desired font family (lowercased).
        weight: Desired font weight (100-900).
        want_italic: Whether italic is desired.

    Returns:
        A score where lower is better.
    """
    family_score = 0.0 if entry.family == family else 1.0
    style_score = 0.0 if entry.is_italic == want_italic else 1.0
    weight_score = abs(entry.weight - weight) / 900
    return family_score * 10 + style_score + weight_score


def find_font(
    family: str,
    weight: momapy.drawing.FontWeight | int,
    style: momapy.drawing.FontStyle,
) -> str | None:
    """Find the best matching font file path.

    Args:
        family: Font family name.
        weight: Font weight as FontWeight enum or int (100-900).
        style: Font style.

    Returns:
        Path to the best matching font file, or None if no fonts found.
    """
    cache_key = (family, weight, style)
    cached = _query_cache.get(cache_key)
    if cached is not None or cache_key in _query_cache:
        return cached
    font_cache = _get_or_make_font_cache()
    if not font_cache:
        _query_cache[cache_key] = None
        return None
    if isinstance(weight, momapy.drawing.FontWeight):
        weight = _FONT_WEIGHT_MAP[weight]
    want_italic = style in (
        momapy.drawing.FontStyle.ITALIC,
        momapy.drawing.FontStyle.OBLIQUE,
    )
    family_lower = family.lower()
    best_font_entry = min(
        font_cache,
        key=lambda font_entry: _score_font(
            font_entry, family_lower, weight, want_italic
        ),
    )
    result = best_font_entry.path
    _query_cache[cache_key] = result
    return result



"""Tests for momapy.utils module."""
import pytest
import momapy.utils


class TestSurjectionDict:
    """Tests for SurjectionDict class."""

    def test_init_empty(self):
        """Test creating an empty SurjectionDict."""
        d = momapy.utils.SurjectionDict()
        assert len(d) == 0
        assert len(d.inverse) == 0

    def test_init_with_items(self):
        """Test creating SurjectionDict with initial items."""
        d = momapy.utils.SurjectionDict({'a': 1, 'b': 2, 'c': 1})
        assert d['a'] == 1
        assert d['b'] == 2
        assert d['c'] == 1
        assert 1 in d.inverse
        assert set(d.inverse[1]) == {'a', 'c'}

    def test_setitem(self):
        """Test setting items."""
        d = momapy.utils.SurjectionDict()
        d['key1'] = 'value1'
        assert d['key1'] == 'value1'
        assert 'value1' in d.inverse

    def test_delitem(self):
        """Test deleting items."""
        d = momapy.utils.SurjectionDict({'a': 1, 'b': 2})
        del d['a']
        assert 'a' not in d
        assert 1 not in d.inverse

    def test_inverse_property(self):
        """Test inverse property."""
        d = momapy.utils.SurjectionDict({'a': 1, 'b': 1, 'c': 2})
        inverse = d.inverse
        assert 1 in inverse
        assert set(inverse[1]) == {'a', 'b'}
        assert set(inverse[2]) == {'c'}


class TestFrozenSurjectionDict:
    """Tests for FrozenSurjectionDict class."""

    def test_init_empty(self):
        """Test creating an empty FrozenSurjectionDict."""
        d = momapy.utils.FrozenSurjectionDict()
        assert len(d) == 0
        assert len(d.inverse) == 0

    def test_init_with_items(self):
        """Test creating FrozenSurjectionDict with items."""
        d = momapy.utils.FrozenSurjectionDict({'a': 1, 'b': 2, 'c': 1})
        assert d['a'] == 1
        assert d['b'] == 2
        assert set(d.inverse[1]) == {'a', 'c'}

    def test_immutable(self):
        """Test that FrozenSurjectionDict is immutable."""
        d = momapy.utils.FrozenSurjectionDict({'a': 1})
        with pytest.raises(Exception):  # frozendict raises TypeError or similar
            d['b'] = 2


def test_get_element_from_collection():
    """Test get_element_from_collection function."""
    collection = [1, 2, 3, 4]
    assert momapy.utils.get_element_from_collection(2, collection) == 2
    assert momapy.utils.get_element_from_collection(5, collection) is None


def test_get_or_return_element_from_collection():
    """Test get_or_return_element_from_collection function."""
    collection = [1, 2, 3, 4]
    assert momapy.utils.get_or_return_element_from_collection(2, collection) == 2
    assert momapy.utils.get_or_return_element_from_collection(5, collection) == 5


def test_add_or_replace_element_in_set():
    """Test add_or_replace_element_in_set function."""
    s = {1, 2, 3}
    result = momapy.utils.add_or_replace_element_in_set(4, s)
    assert result == 4
    assert 4 in s

    # Test with existing element
    result = momapy.utils.add_or_replace_element_in_set(2, s)
    assert result == 2
    assert 2 in s


def test_make_uuid4_as_str():
    """Test make_uuid4_as_str function."""
    uuid1 = momapy.utils.make_uuid4_as_str()
    uuid2 = momapy.utils.make_uuid4_as_str()

    assert isinstance(uuid1, str)
    assert isinstance(uuid2, str)
    assert uuid1 != uuid2
    assert len(uuid1) == 36  # Standard UUID string length
